import os
from typing import Protocol


def _validate_messages(messages: list[dict[str, str]]):
    """
    Validate that the input is a list of dictionaries with "role" and "content" keys.

    Parameters
    ----------
    messages: list[dict[str, str]]
        The messages formatted according to the OpenAI standard. Each message in messages is a dictionary with "role" and "content" keys.
    """
    if not isinstance(messages, list):
        raise TypeError(
            f"messages must be a list, got {type(messages)} instead."
        )

    if not all(isinstance(message, dict) for message in messages):
        raise TypeError("messages must be a list of dictionaries.")

    if not all(
        "role" in message and "content" in message for message in messages
    ):
        raise ValueError(
            'messages must be a list of dictionaries with "role" and "content" keys.'
        )

    if not all(isinstance(message["role"], str) for message in messages):
        raise TypeError("All roles in messages must be strings.")

    if not all(isinstance(message["content"], str) for message in messages):
        raise TypeError("All content in messages must be strings.")


class ClientWrapper(Protocol):
    def __call__(
        self,
        messages: list[dict[str, str]],
    ) -> str:
        ...


class OpenAIWrapper:
    """
    Wrapper for calls to OpenAI's API.

    Attributes
    ----------
    model_name : str
        The model to use. Defaults to "gpt-3.5-turbo".
    api_key : str, optional
        The OpenAI API key to use. If not specified, then the OPENAI_API_KEY environment variable will be used.
    seed : int, optional
        An optional seed can be provided to GPT to get deterministic results.
    total_prompt_tokens : int
        A total count of tokens used for prompt inputs.
    total_completion_tokens : int
        A total count of tokens used to generate responses.
    """

    def __init__(
        self,
        model_name: str,
        api_key: str | None = None,
        seed: int | None = None,
    ):
        """
        Wrapper for calls to OpenAI's API.

        Parameters
        ----------
        model_name : str
            The model to use (e.g. "gpt-3.5-turbo").
        api_key : str, optional
            The OpenAI API key to use. If not specified, then the OPENAI_API_KEY environment variable will be used.
        seed : int, optional
            An optional seed can be provided to GPT to get deterministic results.
        """

        from openai import OpenAI

        if api_key is None:
            self.client = OpenAI()
        else:
            self.client = OpenAI(api_key=api_key)

        self.model_name = model_name
        self.seed = seed

        # logs
        self.total_prompt_tokens = 0
        self.total_completion_tokens = 0

    def __call__(
        self,
        messages: list[dict[str, str]],
    ) -> str:
        """
        Call to the API.

        Parameters
        ----------
        messages: list[dict[str, str]]
            The messages formatted according to the OpenAI standard. Each message in messages is a dictionary with "role" and "content" keys.

        Returns
        -------
        str
            The response from the API.
        """
        _validate_messages(messages=messages)

        openai_response = self.client.chat.completions.create(
            model=self.model_name,
            messages=messages,  # type: ignore - mistralai issue
            seed=self.seed,
        )

        response = openai_response.choices[0].message.content
        if openai_response.usage is not None:
            self.total_prompt_tokens += openai_response.usage.prompt_tokens
            self.total_completion_tokens += (
                openai_response.usage.completion_tokens
            )
        finish_reason = openai_response.choices[
            0
        ].finish_reason  # Enum: "stop" "length" "content_filter" "tool_calls" "function_call"

        if finish_reason == "length":
            raise ValueError(
                "OpenAI response reached max token limit. Resulting evaluation is likely invalid or of low quality."
            )
        elif finish_reason == "content_filter":
            raise ValueError(
                "OpenAI response was flagged by content filter. Resulting evaluation is likely invalid or of low quality."
            )

        if response is None:
            response = ""
        return response


class MistralWrapper:
    """
    Wrapper for calls to Mistral's API.

    Attributes
    ----------
    api_key : str, optional
        The Mistral API key to use. If not specified, then the MISTRAL_API_KEY environment variable will be used.
    model_name : str
        The model to use. Defaults to "mistral-small-latest".
    """

    def __init__(
        self,
        model_name: str,
        api_key: str | None = None,
    ):
        """
        Creates an instance of the Mistral interface.

        Parameters
        ----------
        model_name : str
            The model to use (e.g. "mistral-small-latest").
        api_key : str, optional
            The Mistral API key to use. If not specified, then the MISTRAL_API_KEY environment variable will be used.
        """

        from mistralai import Mistral

        if api_key is None:
            self.client = Mistral(api_key=os.getenv("MISTRAL_API_KEY"))
        else:
            self.client = Mistral(api_key=api_key)

        self.model_name = model_name

    def __call__(
        self,
        messages: list[dict[str, str]],
    ) -> str:
        """
        Call to the API.

        Parameters
        ----------
        messages: list[dict[str, str]]
            The messages formatted according to the OpenAI standard. Each message in messages is a dictionary with "role" and "content" keys.

        Returns
        -------
        str
            The response from the API.
        """
        _validate_messages(messages)

        mistral_response = self.client.chat.complete(
            model=self.model_name,
            messages=messages,  # type: ignore - mistral complaining about native types
        )
        if (
            mistral_response is None
            or mistral_response.choices is None
            or mistral_response.choices[0].message is None
            or mistral_response.choices[0].message.content is None
        ):
            return ""

        response = mistral_response.choices[0].message.content

        finish_reason = mistral_response.choices[
            0
        ].finish_reason  # Enum: "stop" "length" "model_length" "error" "tool_calls"

        if finish_reason == "length":
            raise ValueError(
                "Mistral response reached max token limit. Resulting evaluation is likely invalid or of low quality."
            )

        if not isinstance(response, str):
            raise TypeError("Mistral AI response was not a string.")

        return response



import numpy as np
import pyarrow as pa
from numpy.typing import NDArray
from tqdm import tqdm

from valor_lite.cache import FileCacheWriter, MemoryCacheWriter
from valor_lite.object_detection.annotation import (
    Bitmask,
    BoundingBox,
    Detection,
    Polygon,
)
from valor_lite.object_detection.computation import (
    EPSILON,
    compute_bbox_iou,
    compute_bitmask_iou,
    compute_polygon_iou,
)
from valor_lite.object_detection.evaluator import Builder


class Loader(Builder):
    def __init__(
        self,
        detailed_writer: MemoryCacheWriter | FileCacheWriter,
        ranked_writer: MemoryCacheWriter | FileCacheWriter,
        metadata_fields: list[tuple[str, str | pa.DataType]] | None = None,
    ):
        super().__init__(
            detailed_writer=detailed_writer,
            ranked_writer=ranked_writer,
            metadata_fields=metadata_fields,
        )

        # internal state
        self._labels = {}
        self._datum_count = 0
        self._groundtruth_count = 0
        self._prediction_count = 0

    def _add_label(self, value: str) -> int:
        """Add a label to the index mapping."""
        idx = self._labels.get(value, None)
        if idx is None:
            idx = len(self._labels)
            self._labels[value] = idx
        return idx

    def _add_data(
        self,
        detections: list[Detection],
        detection_ious: list[NDArray[np.float64]],
        show_progress: bool = False,
    ):
        """Adds detections to the cache."""
        disable_tqdm = not show_progress
        for detection, ious in tqdm(
            zip(detections, detection_ious), disable=disable_tqdm
        ):
            # cache labels and annotation pairs
            datum_idx = self._datum_count
            datum_metadata = detection.metadata if detection.metadata else {}
            pairs = []
            if detection.groundtruths:
                for gidx, gann in enumerate(detection.groundtruths):
                    gt_id = self._groundtruth_count + gidx
                    glabel = gann.labels[0]
                    glabel_idx = self._add_label(gann.labels[0])
                    gann_metadata = gann.metadata if gann.metadata else {}
                    if (ious[:, gidx] < EPSILON).all():
                        pairs.append(
                            {
                                # metadata
                                **datum_metadata,
                                **gann_metadata,
                                # datum
                                "datum_uid": detection.uid,
                                "datum_id": datum_idx,
                                # groundtruth
                                "gt_uid": gann.uid,
                                "gt_id": gt_id,
                                "gt_label": glabel,
                                "gt_label_id": glabel_idx,
                                # prediction
                                "pd_uid": None,
                                "pd_id": -1,
                                "pd_label": None,
                                "pd_label_id": -1,
                                "pd_score": -1,
                                # pair
                                "iou": 0.0,
                            }
                        )
                    for pidx, pann in enumerate(detection.predictions):
                        pann_id = self._prediction_count + pidx
                        pann_metadata = pann.metadata if pann.metadata else {}
                        if (ious[pidx, :] < EPSILON).all():
                            pairs.extend(
                                [
                                    {
                                        # metadata
                                        **datum_metadata,
                                        **pann_metadata,
                                        # datum
                                        "datum_uid": detection.uid,
                                        "datum_id": datum_idx,
                                        # groundtruth
                                        "gt_uid": None,
                                        "gt_id": -1,
                                        "gt_label": None,
                                        "gt_label_id": -1,
                                        # prediction
                                        "pd_uid": pann.uid,
                                        "pd_id": pann_id,
                                        "pd_label": plabel,
                                        "pd_label_id": self._add_label(plabel),
                                        "pd_score": float(pscore),
                                        # pair
                                        "iou": 0.0,
                                    }
                                    for plabel, pscore in zip(
                                        pann.labels, pann.scores
                                    )
                                ]
                            )
                        if ious[pidx, gidx] >= EPSILON:
                            pairs.extend(
                                [
                                    {
                                        # metadata
                                        **datum_metadata,
                                        **gann_metadata,
                                        **pann_metadata,
                                        # datum
                                        "datum_uid": detection.uid,
                                        "datum_id": datum_idx,
                                        # groundtruth
                                        "gt_uid": gann.uid,
                                        "gt_id": gt_id,
                                        "gt_label": glabel,
                                        "gt_label_id": self._add_label(glabel),
                                        # prediction
                                        "pd_uid": pann.uid,
                                        "pd_id": pann_id,
                                        "pd_label": plabel,
                                        "pd_label_id": self._add_label(plabel),
                                        "pd_score": float(pscore),
                                        # pair
                                        "iou": float(ious[pidx, gidx]),
                                    }
                                    for glabel in gann.labels
                                    for plabel, pscore in zip(
                                        pann.labels, pann.scores
                                    )
                                ]
                            )
            elif detection.predictions:
                for pidx, pann in enumerate(detection.predictions):
                    pann_id = self._prediction_count + pidx
                    pann_metadata = pann.metadata if pann.metadata else {}
                    pairs.extend(
                        [
                            {
                                # metadata
                                **datum_metadata,
                                **pann_metadata,
                                # datum
                                "datum_uid": detection.uid,
                                "datum_id": datum_idx,
                                # groundtruth
                                "gt_uid": None,
                                "gt_id": -1,
                                "gt_label": None,
                                "gt_label_id": -1,
                                # prediction
                                "pd_uid": pann.uid,
                                "pd_id": pann_id,
                                "pd_label": plabel,
                                "pd_label_id": self._add_label(plabel),
                                "pd_score": float(pscore),
                                # pair
                                "iou": 0.0,
                            }
                            for plabel, pscore in zip(pann.labels, pann.scores)
                        ]
                    )

            self._datum_count += 1
            self._groundtruth_count += len(detection.groundtruths)
            self._prediction_count += len(detection.predictions)

            self._detailed_writer.write_rows(pairs)

    def add_bounding_boxes(
        self,
        detections: list[Detection[BoundingBox]],
        show_progress: bool = False,
    ):
        """
        Adds bounding box detections to the cache.

        Parameters
        ----------
        detections : list[Detection]
            A list of Detection objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """
        ious = [
            compute_bbox_iou(
                np.array(
                    [
                        [gt.extrema, pd.extrema]
                        for pd in detection.predictions
                        for gt in detection.groundtruths
                    ],
                    dtype=np.float64,
                )
            ).reshape(len(detection.predictions), len(detection.groundtruths))
            for detection in detections
        ]
        return self._add_data(
            detections=detections,
            detection_ious=ious,
            show_progress=show_progress,
        )

    def add_polygons(
        self,
        detections: list[Detection[Polygon]],
        show_progress: bool = False,
    ):
        """
        Adds polygon detections to the cache.

        Parameters
        ----------
        detections : list[Detection]
            A list of Detection objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """
        ious = [
            compute_polygon_iou(
                np.array(
                    [
                        [gt.shape, pd.shape]
                        for pd in detection.predictions
                        for gt in detection.groundtruths
                    ]
                )
            ).reshape(len(detection.predictions), len(detection.groundtruths))
            for detection in detections
        ]
        return self._add_data(
            detections=detections,
            detection_ious=ious,
            show_progress=show_progress,
        )

    def add_bitmasks(
        self,
        detections: list[Detection[Bitmask]],
        show_progress: bool = False,
    ):
        """
        Adds bitmask detections to the cache.

        Parameters
        ----------
        detections : list[Detection]
            A list of Detection objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """
        ious = [
            compute_bitmask_iou(
                np.array(
                    [
                        [gt.mask, pd.mask]
                        for pd in detection.predictions
                        for gt in detection.groundtruths
                    ]
                )
            ).reshape(len(detection.predictions), len(detection.groundtruths))
            for detection in detections
        ]
        return self._add_data(
            detections=detections,
            detection_ious=ious,
            show_progress=show_progress,
        )



from __future__ import annotations

import json
from pathlib import Path

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

from valor_lite.cache import (
    FileCacheReader,
    FileCacheWriter,
    MemoryCacheReader,
    MemoryCacheWriter,
    compute,
)
from valor_lite.classification.computation import (
    compute_accuracy,
    compute_confusion_matrix,
    compute_counts,
    compute_f1_score,
    compute_pair_classifications,
    compute_precision,
    compute_recall,
    compute_rocauc,
)
from valor_lite.classification.metric import Metric, MetricType
from valor_lite.classification.shared import (
    EvaluatorInfo,
    decode_metadata_fields,
    encode_metadata_fields,
    extract_counts,
    extract_groundtruth_count_per_label,
    extract_labels,
    generate_cache_path,
    generate_intermediate_cache_path,
    generate_intermediate_schema,
    generate_metadata_path,
    generate_roc_curve_cache_path,
    generate_roc_curve_schema,
    generate_schema,
)
from valor_lite.classification.utilities import (
    create_empty_confusion_matrix_with_examples,
    create_mapping,
    unpack_confusion_matrix,
    unpack_confusion_matrix_with_examples,
    unpack_examples,
    unpack_precision_recall,
    unpack_rocauc,
)
from valor_lite.exceptions import EmptyCacheError


class Builder:
    def __init__(
        self,
        writer: MemoryCacheWriter | FileCacheWriter,
        roc_curve_writer: MemoryCacheWriter | FileCacheWriter,
        intermediate_writer: MemoryCacheWriter | FileCacheWriter,
        metadata_fields: list[tuple[str, str | pa.DataType]] | None = None,
    ):
        self._writer = writer
        self._roc_curve_writer = roc_curve_writer
        self._intermediate_writer = intermediate_writer
        self._metadata_fields = metadata_fields

    @classmethod
    def in_memory(
        cls,
        batch_size: int = 10_000,
        metadata_fields: list[tuple[str, str | pa.DataType]] | None = None,
    ):
        """
        Create an in-memory evaluator cache.

        Parameters
        ----------
        batch_size : int, default=10_000
            The target number of rows to buffer before writing to the cache. Defaults to 10_000.
        metadata_fields : list[tuple[str, str | pa.DataType]], optional
            Optional metadata field definitions.
        """
        writer = MemoryCacheWriter.create(
            schema=generate_schema(metadata_fields),
            batch_size=batch_size,
        )
        intermediate_writer = MemoryCacheWriter.create(
            schema=generate_intermediate_schema(),
            batch_size=batch_size,
        )
        roc_curve_writer = MemoryCacheWriter.create(
            schema=generate_roc_curve_schema(),
            batch_size=batch_size,
        )
        return cls(
            writer=writer,
            roc_curve_writer=roc_curve_writer,
            intermediate_writer=intermediate_writer,
            metadata_fields=metadata_fields,
        )

    @classmethod
    def persistent(
        cls,
        path: str | Path,
        batch_size: int = 10_000,
        rows_per_file: int = 100_000,
        compression: str = "snappy",
        metadata_fields: list[tuple[str, str | pa.DataType]] | None = None,
    ):
        """
        Create a persistent file-based evaluator cache.

        Parameters
        ----------
        path : str | Path
            Where to store file-based cache.
        batch_size : int, default=10_000
            Sets the batch size for writing to file.
        rows_per_file : int, default=100_000
            Sets the maximum number of rows per file. This may be exceeded as files are datum aligned.
        compression : str, default="snappy"
            Sets the pyarrow compression method.
        metadata_fields : list[tuple[str, str | pa.DataType]], optional
            Optionally sets metadata description for use in filtering.
        """
        path = Path(path)

        # create cache
        writer = FileCacheWriter.create(
            path=generate_cache_path(path),
            schema=generate_schema(metadata_fields),
            batch_size=batch_size,
            rows_per_file=rows_per_file,
            compression=compression,
        )
        intermediate_writer = FileCacheWriter.create(
            path=generate_intermediate_cache_path(path),
            schema=generate_intermediate_schema(),
            batch_size=batch_size,
            rows_per_file=rows_per_file,
            compression=compression,
        )
        roc_curve_writer = FileCacheWriter.create(
            path=generate_roc_curve_cache_path(path),
            schema=generate_roc_curve_schema(),
            batch_size=batch_size,
            rows_per_file=rows_per_file,
            compression=compression,
        )

        # write metadatata config
        metadata_path = generate_metadata_path(path)
        with open(metadata_path, "w") as f:
            encoded_types = encode_metadata_fields(metadata_fields)
            json.dump(encoded_types, f, indent=2)

        return cls(
            writer=writer,
            roc_curve_writer=roc_curve_writer,
            intermediate_writer=intermediate_writer,
            metadata_fields=metadata_fields,
        )

    def _create_rocauc_intermediate(
        self,
        reader: MemoryCacheReader | FileCacheReader,
        batch_size: int,
        index_to_label: dict[int, str],
    ):
        n_labels = len(index_to_label)
        compute.sort(
            source=reader,
            sink=self._intermediate_writer,
            batch_size=batch_size,
            sorting=[
                ("pd_score", "descending"),
                # ("pd_label_id", "ascending"),
                ("match", "ascending"),
            ],
            columns=[
                "pd_label_id",
                "pd_score",
                "match",
            ],
        )
        intermediate = self._intermediate_writer.to_reader()

        running_max_fp = np.zeros(n_labels, dtype=np.uint64)
        running_max_tp = np.zeros(n_labels, dtype=np.uint64)
        running_max_scores = np.zeros(n_labels, dtype=np.float64)

        last_pair = np.zeros((n_labels, 2), dtype=np.uint64)
        for tbl in intermediate.iterate_tables():
            pd_label_ids = tbl["pd_label_id"].to_numpy()
            tps = tbl["match"].to_numpy()
            scores = tbl["pd_score"].to_numpy()
            fps = ~tps

            for idx in index_to_label.keys():
                mask = pd_label_ids == idx
                if mask.sum() == 0:
                    continue

                cumulative_fp = np.r_[
                    running_max_fp[idx],
                    np.cumsum(fps[mask]) + running_max_fp[idx],
                ]
                cumulative_tp = np.r_[
                    running_max_tp[idx],
                    np.cumsum(tps[mask]) + running_max_tp[idx],
                ]
                pd_scores = np.r_[running_max_scores[idx], scores[mask]]

                indices = (
                    np.where(
                        np.diff(np.r_[running_max_scores[idx], pd_scores])
                    )[0]
                    - 1
                )

                running_max_fp[idx] = cumulative_fp[-1]
                running_max_tp[idx] = cumulative_tp[-1]
                running_max_scores[idx] = pd_scores[-1]

                for fp, tp in zip(
                    cumulative_fp[indices],
                    cumulative_tp[indices],
                ):
                    last_pair[idx, 0] = fp
                    last_pair[idx, 1] = tp
                    self._roc_curve_writer.write_rows(
                        [
                            {
                                "pd_label_id": idx,
                                "cumulative_fp": fp,
                                "cumulative_tp": tp,
                            }
                        ]
                    )

        # ensure any remaining values are ingested
        for idx in range(n_labels):
            last_fp = last_pair[idx, 0]
            last_tp = last_pair[idx, 1]
            if (
                last_fp != running_max_fp[idx]
                or last_tp != running_max_tp[idx]
            ):
                self._roc_curve_writer.write_rows(
                    [
                        {
                            "pd_label_id": idx,
                            "cumulative_fp": running_max_fp[idx],
                            "cumulative_tp": running_max_tp[idx],
                        }
                    ]
                )

    def finalize(
        self,
        batch_size: int = 1_000,
        index_to_label_override: dict[int, str] | None = None,
    ):
        """
        Performs data finalization and some preprocessing steps.

        Parameters
        ----------
        batch_size : int, default=1_000
            Sets the maximum number of elements read into memory per-file when performing merge sort.
        index_to_label_override : dict[int, str], optional
            Pre-configures label mapping. Used when operating over filtered subsets.

        Returns
        -------
        Evaluator
            A ready-to-use evaluator object.
        """
        self._writer.flush()
        if self._writer.count_rows() == 0:
            raise EmptyCacheError()
        elif self._roc_curve_writer.count_rows() > 0:
            raise RuntimeError("data already finalized")

        # sort in-place and locally
        self._writer.sort_by(
            [
                ("pd_score", "descending"),
                ("datum_id", "ascending"),
                ("gt_label_id", "ascending"),
                ("pd_label_id", "ascending"),
            ]
        )

        # post-process into sorted writer
        reader = self._writer.to_reader()

        # extract labels
        index_to_label = extract_labels(
            reader=reader,
            index_to_label_override=index_to_label_override,
        )

        self._create_rocauc_intermediate(
            reader=reader,
            batch_size=batch_size,
            index_to_label=index_to_label,
        )
        roc_curve_reader = self._roc_curve_writer.to_reader()

        return Evaluator(
            reader=reader,
            roc_curve_reader=roc_curve_reader,
            index_to_label=index_to_label,
            metadata_fields=self._metadata_fields,
        )


class Evaluator:
    def __init__(
        self,
        reader: MemoryCacheReader | FileCacheReader,
        roc_curve_reader: MemoryCacheReader | FileCacheReader,
        index_to_label: dict[int, str],
        metadata_fields: list[tuple[str, str | pa.DataType]] | None = None,
    ):
        self._reader = reader
        self._roc_curve_reader = roc_curve_reader
        self._index_to_label = index_to_label
        self._metadata_fields = metadata_fields

    @property
    def info(self) -> EvaluatorInfo:
        return self.get_info()

    def get_info(
        self,
        datums: pc.Expression | None = None,
    ) -> EvaluatorInfo:
        info = EvaluatorInfo()
        info.metadata_fields = self._metadata_fields
        info.number_of_rows = self._reader.count_rows()
        info.number_of_labels = len(self._index_to_label)
        info.number_of_datums = extract_counts(
            reader=self._reader,
            datums=datums,
        )
        return info

    @classmethod
    def load(
        cls,
        path: str | Path,
        index_to_label_override: dict[int, str] | None = None,
    ):
        """
        Load from an existing classification cache.

        Parameters
        ----------
        path : str | Path
            Path to the existing cache.
        index_to_label_override : dict[int, str], optional
            Option to preset index to label dictionary. Used when loading from filtered caches.
        """

        # validate path
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f"Directory does not exist: {path}")
        elif not path.is_dir():
            raise NotADirectoryError(
                f"Path exists but is not a directory: {path}"
            )

        # load cache
        reader = FileCacheReader.load(generate_cache_path(path))
        roc_curve_reader = FileCacheReader.load(
            generate_roc_curve_cache_path(path)
        )

        # extract labels
        index_to_label = extract_labels(
            reader=reader,
            index_to_label_override=index_to_label_override,
        )

        # read config
        metadata_path = generate_metadata_path(path)
        metadata_fields = None
        with open(metadata_path, "r") as f:
            encoded_types = json.load(f)
            metadata_fields = decode_metadata_fields(encoded_types)

        return cls(
            reader=reader,
            roc_curve_reader=roc_curve_reader,
            index_to_label=index_to_label,
            metadata_fields=metadata_fields,
        )

    def filter(
        self,
        datums: pc.Expression | None = None,
        groundtruths: pc.Expression | None = None,
        predictions: pc.Expression | None = None,
        path: str | Path | None = None,
    ) -> Evaluator:
        """
        Filter evaluator cache.

        Parameters
        ----------
        datums : pc.Expression | None = None
            A filter expression used to filter datums.
        groundtruths : pc.Expression | None = None
            A filter expression used to filter ground truth annotations.
        predictions : pc.Expression | None = None
            A filter expression used to filter predictions.
        path : str | Path, optional
            Where to store the filtered cache if storing on disk.

        Returns
        -------
        Evaluator
            A new evaluator object containing the filtered cache.
        """
        from valor_lite.classification.loader import Loader

        if isinstance(self._reader, FileCacheReader):
            if not path:
                raise ValueError(
                    "expected path to be defined for file-based loader"
                )
            loader = Loader.persistent(
                path=path,
                batch_size=self._reader.batch_size,
                rows_per_file=self._reader.rows_per_file,
                compression=self._reader.compression,
                metadata_fields=self.info.metadata_fields,
            )
        else:
            loader = Loader.in_memory(
                batch_size=self._reader.batch_size,
                metadata_fields=self.info.metadata_fields,
            )

        for tbl in self._reader.iterate_tables(filter=datums):
            columns = (
                "datum_id",
                "gt_label_id",
                "pd_label_id",
            )
            pairs = np.column_stack([tbl[col].to_numpy() for col in columns])

            n_pairs = pairs.shape[0]
            gt_ids = pairs[:, (0, 1)].astype(np.int64)
            pd_ids = pairs[:, (0, 2)].astype(np.int64)

            if groundtruths is not None:
                mask_valid_gt = np.zeros(n_pairs, dtype=np.bool_)
                gt_tbl = tbl.filter(groundtruths)
                gt_pairs = np.column_stack(
                    [
                        gt_tbl[col].to_numpy()
                        for col in ("datum_id", "gt_label_id")
                    ]
                ).astype(np.int64)
                for gt in np.unique(gt_pairs, axis=0):
                    mask_valid_gt |= (gt_ids == gt).all(axis=1)
            else:
                mask_valid_gt = np.ones(n_pairs, dtype=np.bool_)

            if predictions is not None:
                mask_valid_pd = np.zeros(n_pairs, dtype=np.bool_)
                pd_tbl = tbl.filter(predictions)
                pd_pairs = np.column_stack(
                    [
                        pd_tbl[col].to_numpy()
                        for col in ("datum_id", "pd_label_id")
                    ]
                ).astype(np.int64)
                for pd in np.unique(pd_pairs, axis=0):
                    mask_valid_pd |= (pd_ids == pd).all(axis=1)
            else:
                mask_valid_pd = np.ones(n_pairs, dtype=np.bool_)

            mask_valid = mask_valid_gt | mask_valid_pd
            mask_valid_gt &= mask_valid
            mask_valid_pd &= mask_valid

            pairs[~mask_valid_gt, 1] = -1
            pairs[~mask_valid_pd, 2] = -1

            for idx, col in enumerate(columns):
                tbl = tbl.set_column(
                    tbl.schema.names.index(col), col, pa.array(pairs[:, idx])
                )
            # TODO (c.zaloom) - improve write strategy, filtered data could be small
            loader._writer.write_table(tbl)

        return loader.finalize(index_to_label_override=self._index_to_label)

    def iterate_values(self, datums: pc.Expression | None = None):
        columns = [
            "datum_id",
            "gt_label_id",
            "pd_label_id",
            "pd_score",
            "pd_winner",
            "match",
        ]
        for tbl in self._reader.iterate_tables(columns=columns, filter=datums):
            ids = np.column_stack(
                [
                    tbl[col].to_numpy()
                    for col in [
                        "datum_id",
                        "gt_label_id",
                        "pd_label_id",
                    ]
                ]
            )
            scores = tbl["pd_score"].to_numpy()
            winners = tbl["pd_winner"].to_numpy()
            matches = tbl["match"].to_numpy()
            yield ids, scores, winners, matches

    def iterate_values_with_tables(self, datums: pc.Expression | None = None):
        for tbl in self._reader.iterate_tables(filter=datums):
            ids = np.column_stack(
                [
                    tbl[col].to_numpy()
                    for col in [
                        "datum_id",
                        "gt_label_id",
                        "pd_label_id",
                    ]
                ]
            )
            scores = tbl["pd_score"].to_numpy()
            winners = tbl["pd_winner"].to_numpy()
            matches = tbl["match"].to_numpy()
            yield ids, scores, winners, matches, tbl

    def compute_rocauc(
        self, datums: pc.Expression | None = None
    ) -> dict[MetricType, list[Metric]]:
        """
        Compute ROCAUC.

        Parameters
        ----------
        datums : pyarrow.compute.Expression, optional
            Option to filter datums by an expression.

        Returns
        -------
        dict[MetricType, list[Metric]]
            A dictionary mapping MetricType enumerations to lists of computed metrics.
        """
        n_labels = self.info.number_of_labels

        rocauc = np.zeros(n_labels, dtype=np.float64)
        label_counts = extract_groundtruth_count_per_label(
            reader=self._reader,
            number_of_labels=len(self._index_to_label),
            datums=datums,
        )

        prev = np.zeros((n_labels, 2), dtype=np.uint64)
        for array in self._roc_curve_reader.iterate_arrays(
            numeric_columns=[
                "pd_label_id",
                "cumulative_fp",
                "cumulative_tp",
            ],
            filter=datums,
        ):
            rocauc, prev = compute_rocauc(
                rocauc=rocauc,
                array=array,
                gt_count_per_label=label_counts[:, 0],
                pd_count_per_label=label_counts[:, 1],
                n_labels=self.info.number_of_labels,
                prev=prev,
            )

        mean_rocauc = rocauc.mean()

        return unpack_rocauc(
            rocauc=rocauc,
            mean_rocauc=mean_rocauc,
            index_to_label=self._index_to_label,
        )

    def compute_precision_recall(
        self,
        score_thresholds: list[float] = [0.0],
        hardmax: bool = True,
        datums: pc.Expression | None = None,
    ) -> dict[MetricType, list]:
        """
        Performs an evaluation and returns metrics.

        Parameters
        ----------
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        hardmax : bool
            Toggles whether a hardmax is applied to predictions.
        datums : pyarrow.compute.Expression, optional
            Option to filter datums by an expression.

        Returns
        -------
        dict[MetricType, list]
            A dictionary mapping MetricType enumerations to lists of computed metrics.
        """
        if not score_thresholds:
            raise ValueError("At least one score threshold must be passed.")

        n_scores = len(score_thresholds)
        n_datums = self.info.number_of_datums
        n_labels = self.info.number_of_labels

        # intermediates
        counts = np.zeros((n_scores, n_labels, 4), dtype=np.uint64)

        for ids, scores, winners, _ in self.iterate_values(datums=datums):
            batch_counts = compute_counts(
                ids=ids,
                scores=scores,
                winners=winners,
                score_thresholds=np.array(score_thresholds),
                hardmax=hardmax,
                n_labels=n_labels,
            )
            counts += batch_counts

        precision = compute_precision(counts)
        recall = compute_recall(counts)
        f1_score = compute_f1_score(precision, recall)
        accuracy = compute_accuracy(counts, n_datums=n_datums)

        return unpack_precision_recall(
            counts=counts,
            precision=precision,
            recall=recall,
            accuracy=accuracy,
            f1_score=f1_score,
            score_thresholds=score_thresholds,
            hardmax=hardmax,
            index_to_label=self._index_to_label,
        )

    def compute_confusion_matrix(
        self,
        score_thresholds: list[float] = [0.0],
        hardmax: bool = True,
        datums: pc.Expression | None = None,
    ) -> list[Metric]:
        """
        Compute a confusion matrix.

        Parameters
        ----------
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        hardmax : bool
            Toggles whether a hardmax is applied to predictions.
        datums : pyarrow.compute.Expression, optional
            Option to filter datums by an expression.

        Returns
        -------
        list[Metric]
            A list of confusion matrices.
        """
        if not score_thresholds:
            raise ValueError("At least one score threshold must be passed.")

        n_scores = len(score_thresholds)
        n_labels = len(self._index_to_label)
        confusion_matrices = np.zeros(
            (n_scores, n_labels, n_labels), dtype=np.uint64
        )
        unmatched_groundtruths = np.zeros(
            (n_scores, n_labels), dtype=np.uint64
        )
        for ids, scores, winners, matches in self.iterate_values(
            datums=datums
        ):
            (
                mask_tp,
                mask_fp_fn_misclf,
                mask_fn_unmatched,
            ) = compute_pair_classifications(
                ids=ids,
                scores=scores,
                winners=winners,
                score_thresholds=np.array(score_thresholds),
                hardmax=hardmax,
            )

            batch_cm, batch_ugt = compute_confusion_matrix(
                ids=ids,
                mask_tp=mask_tp,
                mask_fp_fn_misclf=mask_fp_fn_misclf,
                mask_fn_unmatched=mask_fn_unmatched,
                score_thresholds=np.array(score_thresholds),
                n_labels=n_labels,
            )
            confusion_matrices += batch_cm
            unmatched_groundtruths += batch_ugt

        return unpack_confusion_matrix(
            confusion_matrices=confusion_matrices,
            unmatched_groundtruths=unmatched_groundtruths,
            index_to_label=self._index_to_label,
            score_thresholds=score_thresholds,
            hardmax=hardmax,
        )

    def compute_examples(
        self,
        score_thresholds: list[float] = [0.0],
        hardmax: bool = True,
        datums: pc.Expression | None = None,
    ) -> list[Metric]:
        """
        Compute examples per datum.

        Note: This function should be used with filtering to reduce response size.

        Parameters
        ----------
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        hardmax : bool
            Toggles whether a hardmax is applied to predictions.
        datums : pyarrow.compute.Expression, optional
            Option to filter datums by an expression.

        Returns
        -------
        list[Metric]
            A list of confusion matrices.
        """
        if not score_thresholds:
            raise ValueError("At least one score threshold must be passed.")

        metrics = []
        for (
            ids,
            scores,
            winners,
            _,
            tbl,
        ) in self.iterate_values_with_tables(datums=datums):
            if ids.size == 0:
                continue

            # extract external identifiers
            index_to_datum_id = create_mapping(
                tbl, ids, 0, "datum_id", "datum_uid"
            )

            (
                mask_tp,
                mask_fp_fn_misclf,
                mask_fn_unmatched,
            ) = compute_pair_classifications(
                ids=ids,
                scores=scores,
                winners=winners,
                score_thresholds=np.array(score_thresholds),
                hardmax=hardmax,
            )

            mask_fn = mask_fp_fn_misclf | mask_fn_unmatched
            mask_fp = mask_fp_fn_misclf

            batch_examples = unpack_examples(
                ids=ids,
                mask_tp=mask_tp,
                mask_fp=mask_fp,
                mask_fn=mask_fn,
                index_to_datum_id=index_to_datum_id,
                score_thresholds=score_thresholds,
                hardmax=hardmax,
                index_to_label=self._index_to_label,
            )
            metrics.extend(batch_examples)

        return metrics

    def compute_confusion_matrix_with_examples(
        self,
        score_thresholds: list[float] = [0.0],
        hardmax: bool = True,
        datums: pc.Expression | None = None,
    ) -> list[Metric]:
        """
        Compute confusion matrix with examples.

        Note: This function should be used with filtering to reduce response size.

        Parameters
        ----------
        metrics : dict[int, Metric]
            Mapping of score threshold index to cached metric.
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        hardmax : bool
            Toggles whether a hardmax is applied to predictions.
        datums : pyarrow.compute.Expression, optional
            Option to filter datums by an expression.

        Returns
        -------
        list[Metric]
            A list of confusion matrices.
        """
        if not score_thresholds:
            raise ValueError("At least one score threshold must be passed.")

        metrics = {
            score_idx: create_empty_confusion_matrix_with_examples(
                score_threshold=score_thresh,
                hardmax=hardmax,
                index_to_label=self._index_to_label,
            )
            for score_idx, score_thresh in enumerate(score_thresholds)
        }
        for (
            ids,
            scores,
            winners,
            _,
            tbl,
        ) in self.iterate_values_with_tables(datums=datums):
            if ids.size == 0:
                continue

            # extract external identifiers
            index_to_datum_id = create_mapping(
                tbl, ids, 0, "datum_id", "datum_uid"
            )

            (
                mask_tp,
                mask_fp_fn_misclf,
                mask_fn_unmatched,
            ) = compute_pair_classifications(
                ids=ids,
                scores=scores,
                winners=winners,
                score_thresholds=np.array(score_thresholds),
                hardmax=hardmax,
            )

            mask_matched = mask_tp | mask_fp_fn_misclf
            mask_unmatched_fn = mask_fn_unmatched

            unpack_confusion_matrix_with_examples(
                metrics=metrics,
                ids=ids,
                scores=scores,
                winners=winners,
                mask_matched=mask_matched,
                mask_unmatched_fn=mask_unmatched_fn,
                index_to_datum_id=index_to_datum_id,
                index_to_label=self._index_to_label,
            )

        return list(metrics.values())



import heapq
import tempfile
from pathlib import Path
from typing import Callable

import pyarrow as pa

from valor_lite.cache.ephemeral import MemoryCacheReader, MemoryCacheWriter
from valor_lite.cache.persistent import FileCacheReader, FileCacheWriter


def _merge(
    source: MemoryCacheReader | FileCacheReader,
    sink: MemoryCacheWriter | FileCacheWriter,
    intermediate_sink: MemoryCacheWriter | FileCacheWriter,
    batch_size: int,
    sorting: list[tuple[str, str]],
    columns: list[str] | None = None,
    table_sort_override: Callable[[pa.Table], pa.Table] | None = None,
):
    """Merge locally sorted cache fragments."""
    for tbl in source.iterate_tables(columns=columns):
        if table_sort_override is not None:
            sorted_tbl = table_sort_override(tbl)
        else:
            sorted_tbl = tbl.sort_by(sorting)
        intermediate_sink.write_table(sorted_tbl)
    intermediate_source = intermediate_sink.to_reader()

    # define merge key
    def create_sort_key(
        batches: list[pa.RecordBatch],
        batch_idx: int,
        row_idx: int,
    ):
        args = [
            -batches[batch_idx][name][row_idx].as_py()
            if direction == "descending"
            else batches[batch_idx][name][row_idx].as_py()
            for name, direction in sorting
        ]
        return (
            *args,
            batch_idx,
            row_idx,
        )

    # merge sorted rows
    heap = []
    batch_iterators = []
    batches = []
    for batch_idx, batch_iter in enumerate(
        intermediate_source.iterate_fragments(batch_size=batch_size)
    ):
        batch_iterators.append(batch_iter)
        batches.append(next(batch_iterators[batch_idx], None))
        if batches[batch_idx] is not None and len(batches[batch_idx]) > 0:
            heap.append(create_sort_key(batches, batch_idx, 0))
    heapq.heapify(heap)

    while heap:
        row = heapq.heappop(heap)
        batch_idx = row[-2]
        row_idx = row[-1]
        row_table = batches[batch_idx].slice(row_idx, 1)
        sink.write_batch(row_table)
        row_idx += 1
        if row_idx < len(batches[batch_idx]):
            heapq.heappush(
                heap,
                create_sort_key(batches, batch_idx, row_idx),
            )
        else:
            batches[batch_idx] = next(batch_iterators[batch_idx], None)
            if batches[batch_idx] is not None and len(batches[batch_idx]) > 0:
                heapq.heappush(
                    heap,
                    create_sort_key(batches, batch_idx, 0),
                )

    sink.flush()


def sort(
    source: MemoryCacheReader | FileCacheReader,
    sink: MemoryCacheWriter | FileCacheWriter,
    batch_size: int,
    sorting: list[tuple[str, str]],
    columns: list[str] | None = None,
    table_sort_override: Callable[[pa.Table], pa.Table] | None = None,
):
    """
    Sort data into new cache.

    Parameters
    ----------
    source : MemoryCacheReader | FileCacheReader
        A read-only cache. If file-based, each file must be locally sorted.
    sink : MemoryCacheWriter | FileCacheWriter
        The cache where sorted data will be written.
    batch_size : int
        Maximum number of rows allowed to be read into memory per cache file.
    sorting : list[tuple[str, str]]
        Sorting arguments in PyArrow format (e.g. [('a', 'ascending'), ('b', 'descending')]).
        Note that only numeric fields are supported.
    columns : list[str], optional
        Option to only read a subset of columns.
    table_sort_override : Callable[[pa.Table], pa.Table], optional
        Option to override sort function for singular cache fragments.
    """

    if source.count_tables() == 1:
        for tbl in source.iterate_tables(columns=columns):
            if table_sort_override is not None:
                sorted_tbl = table_sort_override(tbl)
            else:
                sorted_tbl = tbl.sort_by(sorting)
            sink.write_table(sorted_tbl)
        sink.flush()
        return

    if isinstance(sink, FileCacheWriter):
        with tempfile.TemporaryDirectory() as tmpdir:
            intermediate_sink = FileCacheWriter.create(
                path=Path(tmpdir) / "sorting_intermediate",
                schema=sink.schema,
                batch_size=sink.batch_size,
                rows_per_file=sink.rows_per_file,
                compression=sink.compression,
                delete_if_exists=False,
            )
            _merge(
                source=source,
                sink=sink,
                intermediate_sink=intermediate_sink,
                batch_size=batch_size,
                sorting=sorting,
                columns=columns,
                table_sort_override=table_sort_override,
            )
    else:
        intermediate_sink = MemoryCacheWriter.create(
            schema=sink.schema,
            batch_size=sink.batch_size,
        )
        _merge(
            source=source,
            sink=sink,
            intermediate_sink=intermediate_sink,
            batch_size=batch_size,
            sorting=sorting,
            columns=columns,
            table_sort_override=table_sort_override,
        )



import base64
import glob
import json
import os
from collections.abc import Iterator
from pathlib import Path
from typing import Any

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
import pyarrow.parquet as pq


class FileCache:
    def __init__(
        self,
        path: str | Path,
        schema: pa.Schema,
        batch_size: int,
        rows_per_file: int,
        compression: str,
    ):
        self._path = Path(path)
        self._schema = schema
        self._batch_size = batch_size
        self._rows_per_file = rows_per_file
        self._compression = compression

    @property
    def path(self) -> Path:
        return self._path

    @property
    def schema(self) -> pa.Schema:
        return self._schema

    @property
    def batch_size(self) -> int:
        return self._batch_size

    @property
    def rows_per_file(self) -> int:
        return self._rows_per_file

    @property
    def compression(self) -> str:
        return self._compression

    @staticmethod
    def _generate_config_path(path: str | Path) -> Path:
        """Generate cache configuration path."""
        return Path(path) / ".cfg"

    @staticmethod
    def _encode_schema(schema: pa.Schema) -> str:
        """Encode schema to b64 string."""
        schema_bytes = schema.serialize()
        return base64.b64encode(schema_bytes).decode("utf-8")

    @staticmethod
    def _decode_schema(encoded_schema: str) -> pa.Schema:
        """Decode schema from b64 string."""
        schema_bytes = base64.b64decode(encoded_schema)
        return pa.ipc.read_schema(pa.BufferReader(schema_bytes))

    def count_rows(self) -> int:
        """Count the number of rows in the cache."""
        dataset = ds.dataset(
            source=self._path,
            format="parquet",
        )
        return dataset.count_rows()

    def count_tables(self) -> int:
        """Count the number of files in the cache."""
        return len(self.get_dataset_files())

    def get_files(self) -> list[Path]:
        """
        Retrieve all files.

        Returns
        -------
        list[Path]
            A list of paths to files in the cache.
        """
        if not self._path.exists():
            return []
        files = []
        for entry in os.listdir(self._path):
            full_path = os.path.join(self._path, entry)
            if os.path.isfile(full_path):
                files.append(Path(full_path))
        return files

    def get_dataset_files(self) -> list[Path]:
        """
        Retrieve all dataset files.

        Returns
        -------
        list[Path]
            A list of paths to dataset files in the cache.
        """
        if not self._path.exists():
            return []
        return [
            Path(filepath) for filepath in glob.glob(f"{self._path}/*.parquet")
        ]


class FileCacheReader(FileCache):
    @classmethod
    def load(cls, path: str | Path):
        """
        Load cache from disk.

        Parameters
        ----------
        path : str | Path
            Where the cache is stored.
        """
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f"Directory does not exist: {path}")
        elif not path.is_dir():
            raise NotADirectoryError(
                f"Path exists but is not a directory: {path}"
            )

        def _retrieve(config: dict, key: str):
            if value := config.get(key, None):
                return value
            raise KeyError(
                f"'{key}' is not defined within {cls._generate_config_path(path)}"
            )

        # read configuration file
        cfg_path = cls._generate_config_path(path)
        with open(cfg_path, "r") as f:
            cfg = json.load(f)
            batch_size = _retrieve(cfg, "batch_size")
            rows_per_file = _retrieve(cfg, "rows_per_file")
            compression = _retrieve(cfg, "compression")
            schema = cls._decode_schema(_retrieve(cfg, "schema"))

        return cls(
            schema=schema,
            path=path,
            batch_size=batch_size,
            rows_per_file=rows_per_file,
            compression=compression,
        )

    def iterate_tables(
        self,
        columns: list[str] | None = None,
        filter: pc.Expression | None = None,
    ) -> Iterator[pa.Table]:
        """
        Iterate over tables within the cache.

        Parameters
        ----------
        columns : list[str], optional
            Optionally select columns to be returned.
        filter : pyarrow.compute.Expression, optional
            Optionally filter table before returning.

        Returns
        -------
        Iterator[pa.Table]
        """
        dataset = ds.dataset(
            source=self._path,
            schema=self._schema,
            format="parquet",
        )
        for fragment in dataset.get_fragments():
            yield fragment.to_table(columns=columns, filter=filter)

    def iterate_arrays(
        self,
        numeric_columns: list[str] | None = None,
        filter: pc.Expression | None = None,
    ) -> Iterator[np.ndarray]:
        """
        Iterate over chunks within the cache returning arrays.

        Parameters
        ----------
        numeric_columns : list[str], optional
            Optionally select numeric columns to be returned within an array.
        filter : pyarrow.compute.Expression, optional
            Optionally filter table before returning.

        Returns
        -------
        Iterator[np.ndarray]
        """
        for tbl in self.iterate_tables(
            columns=numeric_columns,
            filter=filter,
        ):
            yield np.column_stack(
                [tbl.column(i).to_numpy() for i in range(tbl.num_columns)]
            )

    def iterate_tables_with_arrays(
        self,
        columns: list[str] | None = None,
        filter: pc.Expression | None = None,
        numeric_columns: list[str] | None = None,
    ) -> Iterator[tuple[pa.Table, np.ndarray]]:
        """
        Iterate over chunks within the cache returning both tables and arrays.

        Parameters
        ----------
        columns : list[str], optional
            Optionally select columns to be returned.
        filter : pyarrow.compute.Expression, optional
            Optionally filter table before returning.
        numeric_columns : list[str], optional
            Optionally select numeric columns to be returned within an array.

        Returns
        -------
        Iterator[tuple[pa.Table, np.ndarray]]

        """
        _columns = set(columns) if columns else set()
        _numeric_columns = set(numeric_columns) if numeric_columns else set()
        columns = list(_columns.union(_numeric_columns))
        for tbl in self.iterate_tables(
            columns=columns,
            filter=filter,
        ):
            table_columns = numeric_columns if numeric_columns else tbl.columns
            yield tbl, np.column_stack(
                [tbl[col].to_numpy() for col in table_columns]
            )

    def iterate_fragments(
        self, batch_size: int
    ) -> Iterator[Iterator[pa.RecordBatch]]:
        """
        Iterate over fragment batch iterators within the file-based cache.

        Parameters
        ----------
        batch_size : int
            Maximum number of rows allowed to be read into memory per cache file.

        Returns
        -------
        Iterator[Iterator[pa.RecordBatch]]
        """
        dataset = ds.dataset(
            source=self._path,
            schema=self._schema,
            format="parquet",
        )
        for fragment in dataset.get_fragments():
            yield fragment.to_batches(batch_size=batch_size)


class FileCacheWriter(FileCache):
    def __init__(
        self,
        path: str | Path,
        schema: pa.Schema,
        batch_size: int,
        rows_per_file: int,
        compression: str,
    ):
        super().__init__(
            path=path,
            schema=schema,
            batch_size=batch_size,
            rows_per_file=rows_per_file,
            compression=compression,
        )

        # internal state
        self._writer = None
        self._buffer = []
        self._count = 0

    @classmethod
    def create(
        cls,
        path: str | Path,
        schema: pa.Schema,
        batch_size: int,
        rows_per_file: int,
        compression: str = "snappy",
        delete_if_exists: bool = False,
    ):
        """
        Create an on-disk cache.

        Parameters
        ----------
        path : str | Path
            Where to write the cache.
        schema : pa.Schema
            Cache schema.
        batch_size : int
            Target batch size when writing chunks.
        rows_per_file : int
            Target number of rows to store per file.
        compression : str, default="snappy"
            Compression method to use when storing on disk.
        delete_if_exists : bool, default=False
            Delete the cache if it already exists.
        """
        path = Path(path)
        if delete_if_exists and path.exists():
            cls.delete(path)
        Path(path).mkdir(parents=True, exist_ok=False)

        # write configuration file
        cfg_path = cls._generate_config_path(path)
        with open(cfg_path, "w") as f:
            cfg = dict(
                batch_size=batch_size,
                rows_per_file=rows_per_file,
                compression=compression,
                schema=cls._encode_schema(schema),
            )
            json.dump(cfg, f, indent=2)

        return cls(
            schema=schema,
            path=path,
            batch_size=batch_size,
            rows_per_file=rows_per_file,
            compression=compression,
        )

    @classmethod
    def delete(cls, path: str | Path):
        """
        Delete a cache at path.

        Parameters
        ----------
        path : str | Path
            Where the cache is stored.
        """
        path = Path(path)
        if not path.exists():
            return

        # delete dataset files
        reader = FileCacheReader.load(path)
        for file in reader.get_dataset_files():
            if file.exists() and file.is_file() and file.suffix == ".parquet":
                file.unlink()

        # delete config file
        cfg_path = cls._generate_config_path(path)
        if cfg_path.exists() and cfg_path.is_file():
            cfg_path.unlink()

        # delete empty cache directory
        path.rmdir()

    def write_rows(
        self,
        rows: list[dict[str, Any]],
    ):
        """
        Write rows to cache.

        Parameters
        ----------
        rows : list[dict[str, Any]]
            A list of rows represented by dictionaries mapping fields to values.
        """
        if not rows:
            return
        batch = pa.RecordBatch.from_pylist(rows, schema=self._schema)
        self.write_batch(batch)

    def write_columns(
        self,
        columns: dict[str, list | np.ndarray | pa.Array],
    ):
        """
        Write columnar data to cache.

        Parameters
        ----------
        columns : dict[str, list | np.ndarray | pa.Array]
            A mapping of columnar field names to list of values.
        """
        if not columns:
            return
        batch = pa.RecordBatch.from_pydict(columns)
        self.write_batch(batch)

    def write_batch(
        self,
        batch: pa.RecordBatch,
    ):
        """
        Write a batch to cache.

        Parameters
        ----------
        batch : pa.RecordBatch
            A batch of columnar data.
        """
        size = batch.num_rows
        if self._buffer:
            size += sum([b.num_rows for b in self._buffer])

        # check size
        if size < self.batch_size and self._count < self.rows_per_file:
            self._buffer.append(batch)
            return

        if self._buffer:
            self._buffer.append(batch)
            batch = pa.concat_batches(self._buffer)
            self._buffer = []

        # write batch
        writer = self._get_or_create_writer()
        writer.write_batch(batch)

        # check file size
        self._count += size
        if self._count >= self.rows_per_file:
            self.flush()

    def write_table(
        self,
        table: pa.Table,
    ):
        """
        Write a table directly to cache.

        Parameters
        ----------
        table : pa.Table
            A populated table.
        """
        self.flush()
        pq.write_table(table, where=self._generate_next_filename())

    def flush(self):
        """Flush the cache buffer."""
        if self._buffer:
            combined_arrays = [
                pa.concat_arrays([b.column(name) for b in self._buffer])
                for name in self._schema.names
            ]
            batch = pa.RecordBatch.from_arrays(
                combined_arrays, schema=self._schema
            )
            writer = self._get_or_create_writer()
            writer.write_batch(batch)
        self._buffer = []
        self._count = 0
        self._close_writer()

    def sort_by(
        self,
        sorting: list[tuple[str, str]],
    ):
        """
        Sort cache files locally and in-place.

        Parameters
        ----------
        sorting : list[tuple[str, str]]
            Sorting arguments in PyArrow format (e.g. [('a', 'ascending'), ('b', 'descending')]).
        """
        self.flush()
        for file in self.get_dataset_files():
            pf = pq.ParquetFile(file)
            tbl = pf.read()
            pf.close()
            sorted_tbl = tbl.sort_by(sorting)
            pq.write_table(sorted_tbl, file)

    def _generate_next_filename(self) -> Path:
        """Generates next dataset filepath."""
        files = self.get_dataset_files()
        if not files:
            next_index = 0
        else:
            next_index = max([int(Path(f).stem) for f in files]) + 1
        return self._path / f"{next_index:06d}.parquet"

    def _get_or_create_writer(self) -> pq.ParquetWriter:
        """Open a new parquet file for writing."""
        if self._writer is not None:
            return self._writer
        self._writer = pq.ParquetWriter(
            where=self._generate_next_filename(),
            schema=self._schema,
            compression=self._compression,
        )
        return self._writer

    def _close_writer(self) -> None:
        """Close the current parquet file."""
        if self._writer is not None:
            self._writer.close()
            self._writer = None

    def __enter__(self):
        """Context manager entry."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit - ensures data is flushed."""
        self.flush()

    def to_reader(self) -> FileCacheReader:
        """Get cache reader."""
        self.flush()
        return FileCacheReader.load(path=self.path)



from dataclasses import dataclass
from enum import Enum

from valor_lite.schemas import BaseMetric


class MetricType(str, Enum):
    AnswerCorrectness = "AnswerCorrectness"
    AnswerRelevance = "AnswerRelevance"
    Bias = "Bias"
    BLEU = "BLEU"
    ContextPrecision = "ContextPrecision"
    ContextRecall = "ContextRecall"
    ContextRelevance = "ContextRelevance"
    Faithfulness = "Faithfulness"
    Hallucination = "Hallucination"
    ROUGE = "ROUGE"
    SummaryCoherence = "SummaryCoherence"
    Toxicity = "Toxicity"


@dataclass
class Metric(BaseMetric):
    """
    Text Generation Metric.

    Attributes
    ----------
    type : str
        The metric type.
    value : int | float | dict
        The metric value.
    parameters : dict[str, Any]
        A dictionary containing metric parameters.
    """

    def __post_init__(self):
        if not isinstance(self.type, str):
            raise TypeError(
                f"Metric type should be of type 'str': {self.type}"
            )
        elif not isinstance(self.value, (int, float, dict)):
            raise TypeError(
                f"Metric value must be of type 'int', 'float' or 'dict': {self.value}"
            )
        elif not isinstance(self.parameters, dict):
            raise TypeError(
                f"Metric parameters must be of type 'dict[str, Any]': {self.parameters}"
            )
        elif not all([isinstance(k, str) for k in self.parameters.keys()]):
            raise TypeError(
                f"Metric parameter dictionary should only have keys with type 'str': {self.parameters}"
            )

    @classmethod
    def error(
        cls,
        error_type: str,
        error_message: str,
        model_name: str,
        retries: int,
    ):
        return cls(
            type="Error",
            value={
                "type": error_type,
                "message": error_message,
            },
            parameters={
                "evaluator": model_name,
                "retries": retries,
            },
        )

    @classmethod
    def answer_correctness(
        cls,
        value: float,
        model_name: str,
        retries: int,
    ):
        """
        Defines an answer correctness metric.

        Parameters
        ----------
        value : float
            The answer correctness score between 0 and 1, with higher values indicating that the answer
            is more correct. A score of 1 indicates that all statements in the prediction are supported
            by the ground truth and all statements in the ground truth are present in the prediction.
        """
        return cls(
            type=MetricType.AnswerCorrectness,
            value=value,
            parameters={
                "evaluator": model_name,
                "retries": retries,
            },
        )

    @classmethod
    def answer_relevance(
        cls,
        value: float,
        model_name: str,
        retries: int,
    ):
        """
        Defines an answer relevance metric.

        Parameters
        ----------
        value : float
            The number of statements in the answer that are relevant to the query divided by the total
            number of statements in the answer.
        """
        return cls(
            type=MetricType.AnswerRelevance,
            value=value,
            parameters={
                "evaluator": model_name,
                "retries": retries,
            },
        )

    @classmethod
    def bleu(
        cls,
        value: float,
        weights: list[float],
    ):
        """
        Defines a BLEU metric.

        Parameters
        ----------
        value : float
            The BLEU score for an individual datapoint.
        weights : list[float]
            The list of weights that the score was calculated with.
        """
        return cls(
            type=MetricType.BLEU,
            value=value,
            parameters={
                "weights": weights,
            },
        )

    @classmethod
    def bias(
        cls,
        value: float,
        model_name: str,
        retries: int,
    ):
        """
        Defines a bias metric.

        Parameters
        ----------
        value : float
            The bias score for a datum. This is a float between 0 and 1, with 1 indicating that all
            opinions in the datum text are biased and 0 indicating that there is no bias.
        """
        return cls(
            type=MetricType.Bias,
            value=value,
            parameters={
                "evaluator": model_name,
                "retries": retries,
            },
        )

    @classmethod
    def context_precision(
        cls,
        value: float,
        model_name: str,
        retries: int,
    ):
        """
        Defines a context precision metric.

        Parameters
        ----------
        value : float
            The context precision score for a datum. This is a float between 0 and 1, with 0 indicating
            that none of the contexts are useful to arrive at the ground truth answer to the query
            and 1 indicating that all contexts are useful to arrive at the ground truth answer to the
            query. The score is more heavily influenced by earlier contexts in the list of contexts
            than later contexts.
        """
        return cls(
            type=MetricType.ContextPrecision,
            value=value,
            parameters={
                "evaluator": model_name,
                "retries": retries,
            },
        )

    @classmethod
    def context_recall(
        cls,
        value: float,
        model_name: str,
        retries: int,
    ):
        """
        Defines a context recall metric.

        Parameters
        ----------
        value : float
            The context recall score for a datum. This is a float between 0 and 1, with 1 indicating
            that all ground truth statements are attributable to the context list.
        """
        return cls(
            type=MetricType.ContextRecall,
            value=value,
            parameters={
                "evaluator": model_name,
                "retries": retries,
            },
        )

    @classmethod
    def context_relevance(
        cls,
        value: float,
        model_name: str,
        retries: int,
    ):
        """
        Defines a context relevance metric.

        Parameters
        ----------
        value : float
            The context relevance score for a datum. This is a float between 0 and 1, with 0 indicating
            that none of the contexts are relevant and 1 indicating that all of the contexts are relevant.
        """
        return cls(
            type=MetricType.ContextRelevance,
            value=value,
            parameters={
                "evaluator": model_name,
                "retries": retries,
            },
        )

    @classmethod
    def faithfulness(
        cls,
        value: float,
        model_name: str,
        retries: int,
    ):
        """
        Defines a faithfulness metric.

        Parameters
        ----------
        value : float
            The faithfulness score for a datum. This is a float between 0 and 1, with 1 indicating that
            all claims in the text are implied by the contexts.
        """
        return cls(
            type=MetricType.Faithfulness,
            value=value,
            parameters={
                "evaluator": model_name,
                "retries": retries,
            },
        )

    @classmethod
    def hallucination(
        cls,
        value: float,
        model_name: str,
        retries: int,
    ):
        """
        Defines a hallucination metric.

        Parameters
        ----------
        value : float
            The hallucination score for a datum. This is a float between 0 and 1, with 1 indicating that
            all contexts are contradicted by the text.
        """
        return cls(
            type=MetricType.Hallucination,
            value=value,
            parameters={
                "evaluator": model_name,
                "retries": retries,
            },
        )

    @classmethod
    def rouge(
        cls,
        value: float,
        rouge_type: str,
        use_stemmer: bool,
    ):
        """
        Defines a ROUGE metric.

        Parameters
        ----------
        value : float
            A ROUGE score.
        rouge_type : ROUGEType
            The ROUGE variation used to compute the value. `rouge1` is unigram-based scoring, `rouge2` is bigram-based
            scoring, `rougeL` is scoring based on sentences (i.e., splitting on "." and ignoring "\n"), and `rougeLsum`
            is scoring based on splitting the text using "\n".
        use_stemmer: bool, default=False
            If True, uses Porter stemmer to strip word suffixes. Defaults to False.
        """
        return cls(
            type=MetricType.ROUGE,
            value=value,
            parameters={
                "rouge_type": rouge_type,
                "use_stemmer": use_stemmer,
            },
        )

    @classmethod
    def summary_coherence(
        cls,
        value: int,
        model_name: str,
        retries: int,
    ):
        """
        Defines a summary coherence metric.

        Parameters
        ----------
        value : int
            The summary coherence score for a datum. This is an integer with 1 being the lowest summary coherence
            and 5 the highest summary coherence.
        """
        return cls(
            type=MetricType.SummaryCoherence,
            value=value,
            parameters={
                "evaluator": model_name,
                "retries": retries,
            },
        )

    @classmethod
    def toxicity(
        cls,
        value: float,
        model_name: str,
        retries: int,
    ):
        """
        Defines a toxicity metric.

        Parameters
        ----------
        value : float
            The toxicity score for a datum. This is a value between 0 and 1, with 1 indicating that all opinions
            in the datum text are toxic and 0 indicating that there is no toxicity.
        """
        return cls(
            type=MetricType.Toxicity,
            value=value,
            parameters={
                "evaluator": model_name,
                "retries": retries,
            },
        )



from .annotation import Bitmask, Segmentation
from .evaluator import Builder, Evaluator, EvaluatorInfo
from .loader import Loader
from .metric import Metric, MetricType

__all__ = [
    "Builder",
    "Loader",
    "Evaluator",
    "Segmentation",
    "Bitmask",
    "Metric",
    "MetricType",
    "EvaluatorInfo",
]



import numpy as np
import pyarrow as pa
from tqdm import tqdm

from valor_lite.cache import FileCacheWriter, MemoryCacheWriter
from valor_lite.semantic_segmentation.annotation import Segmentation
from valor_lite.semantic_segmentation.computation import compute_intermediates
from valor_lite.semantic_segmentation.evaluator import Builder


class Loader(Builder):
    def __init__(
        self,
        writer: MemoryCacheWriter | FileCacheWriter,
        metadata_fields: list[tuple[str, str | pa.DataType]] | None = None,
    ):
        super().__init__(
            writer=writer,
            metadata_fields=metadata_fields,
        )

        # internal state
        self._labels: dict[str, int] = {}
        self._index_to_label: dict[int, str] = {}
        self._datum_count = 0

    def _add_label(self, value: str) -> int:
        idx = self._labels.get(value, None)
        if idx is None:
            idx = len(self._labels)
            self._labels[value] = idx
            self._index_to_label[idx] = value
        return idx

    def add_data(
        self,
        segmentations: list[Segmentation],
        show_progress: bool = False,
    ):
        """
        Adds segmentations to the cache.

        Parameters
        ----------
        segmentations : list[Segmentation]
            A list of Segmentation objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """

        disable_tqdm = not show_progress
        for segmentation in tqdm(segmentations, disable=disable_tqdm):

            groundtruth_labels = -1 * np.ones(
                len(segmentation.groundtruths), dtype=np.int64
            )
            for idx, groundtruth in enumerate(segmentation.groundtruths):
                label_idx = self._add_label(groundtruth.label)
                groundtruth_labels[idx] = label_idx

            prediction_labels = -1 * np.ones(
                len(segmentation.predictions), dtype=np.int64
            )
            for idx, prediction in enumerate(segmentation.predictions):
                label_idx = self._add_label(prediction.label)
                prediction_labels[idx] = label_idx

            if segmentation.groundtruths:
                combined_groundtruths = np.stack(
                    [
                        groundtruth.mask.flatten()
                        for groundtruth in segmentation.groundtruths
                    ],
                    axis=0,
                )
            else:
                combined_groundtruths = np.zeros(
                    (1, segmentation.shape[0] * segmentation.shape[1]),
                    dtype=np.bool_,
                )

            if segmentation.predictions:
                combined_predictions = np.stack(
                    [
                        prediction.mask.flatten()
                        for prediction in segmentation.predictions
                    ],
                    axis=0,
                )
            else:
                combined_predictions = np.zeros(
                    (1, segmentation.shape[0] * segmentation.shape[1]),
                    dtype=np.bool_,
                )

            n_labels = len(self._labels)
            counts = compute_intermediates(
                groundtruths=combined_groundtruths,
                predictions=combined_predictions,
                groundtruth_labels=groundtruth_labels,
                prediction_labels=prediction_labels,
                n_labels=n_labels,
            )

            # prepare metadata
            datum_metadata = (
                segmentation.metadata if segmentation.metadata else {}
            )
            gt_metadata = {
                self._labels[gt.label]: gt.metadata
                for gt in segmentation.groundtruths
                if gt.metadata
            }
            pd_metadata = {
                self._labels[pd.label]: pd.metadata
                for pd in segmentation.predictions
                if pd.metadata
            }

            # cache formatting
            rows = []
            for idx in range(n_labels):
                label = self._index_to_label[idx]
                for pidx in range(n_labels):
                    # write non-zero intersections to cache
                    if counts[idx + 1, pidx + 1] > 0:
                        plabel = self._index_to_label[pidx]
                        rows.append(
                            {
                                # metadata
                                **datum_metadata,
                                **gt_metadata.get(idx, {}),
                                **pd_metadata.get(pidx, {}),
                                # datum
                                "datum_uid": segmentation.uid,
                                "datum_id": self._datum_count,
                                # groundtruth
                                "gt_label": label,
                                "gt_label_id": idx,
                                # prediction
                                "pd_label": plabel,
                                "pd_label_id": pidx,
                                # pair
                                "count": counts[idx + 1, pidx + 1],
                            }
                        )
                # write all unmatched to preserve labels
                rows.extend(
                    [
                        {
                            # metadata
                            **datum_metadata,
                            **gt_metadata.get(idx, {}),
                            # datum
                            "datum_uid": segmentation.uid,
                            "datum_id": self._datum_count,
                            # groundtruth
                            "gt_label": label,
                            "gt_label_id": idx,
                            # prediction
                            "pd_label": None,
                            "pd_label_id": -1,
                            # pair
                            "count": counts[idx + 1, 0],
                        },
                        {
                            # metadata
                            **datum_metadata,
                            **gt_metadata.get(idx, {}),
                            **pd_metadata.get(idx, {}),
                            # datum
                            "datum_uid": segmentation.uid,
                            "datum_id": self._datum_count,
                            # groundtruth
                            "gt_label": None,
                            "gt_label_id": -1,
                            # prediction
                            "pd_label": label,
                            "pd_label_id": idx,
                            # pair
                            "count": counts[0, idx + 1],
                        },
                    ]
                )
            rows.append(
                {
                    # metadata
                    **datum_metadata,
                    # datum
                    "datum_uid": segmentation.uid,
                    "datum_id": self._datum_count,
                    # groundtruth
                    "gt_label": None,
                    "gt_label_id": -1,
                    # prediction
                    "pd_label": None,
                    "pd_label_id": -1,
                    # pair
                    "count": counts[0, 0],
                }
            )
            self._writer.write_rows(rows)

            # update datum count
            self._datum_count += 1



from enum import IntFlag

import numpy as np
from numpy.typing import NDArray

import valor_lite.classification.numpy_compatibility as npc


def compute_rocauc(
    rocauc: NDArray[np.float64],
    array: NDArray[np.float64],
    gt_count_per_label: NDArray[np.uint64],
    pd_count_per_label: NDArray[np.uint64],
    n_labels: int,
    prev: NDArray[np.uint64],
) -> tuple[NDArray[np.float64], NDArray[np.uint64]]:
    """
    Compute ROCAUC.

    Parameters
    ----------
    rocauc : NDArray[np.float64]
        The running ROCAUC.
    array : NDArray[np.float64]
        An sorted array of ROCAUC intermediate values with shape (n_pairs, 3).
            Index 0 - Prediction Label Index
            Index 1 - Cumulative FP
            Index 2 - Cumulative TP
    gt_count_per_label : NDArray[np.uint64]
        The number of ground truth occurences per label.
    pd_count_per_label : NDArray[np.uint64]
        The number of prediction occurences per label.
    n_labels : int
        The total number of unqiue labels.
    prev : NDArray[np.uint64]
        The previous cumulative sum for FP's and TP's. Used as intermediate in chunking operations.

    Returns
    -------
    NDArray[np.float64]
        ROCAUC.
    NDArray[np.uint64]
        The previous cumulative sum for FP's. Used as intermediate in chunking operations.
    NDArray[np.uint64]
        The previous cumulative sum for TP's. Used as intermediate in chunking operations.
    """
    pd_labels = array[:, 0]
    cumulative_fp = array[:, 1]
    cumulative_tp = array[:, 2]

    positive_count = gt_count_per_label
    negative_count = pd_count_per_label - gt_count_per_label

    for label_idx in range(n_labels):
        mask_pds = pd_labels == label_idx
        n_masked_pds = mask_pds.sum()
        if pd_count_per_label[label_idx] == 0 or n_masked_pds == 0:
            continue

        fps = cumulative_fp[mask_pds]
        tps = cumulative_tp[mask_pds]
        if prev[label_idx, 0] > 0 or prev[label_idx, 1] > 0:
            fps = np.r_[prev[label_idx, 0], fps]
            tps = np.r_[prev[label_idx, 1], tps]

        prev[label_idx, 0] = fps[-1]
        prev[label_idx, 1] = tps[-1]

        if fps.size == 1:
            continue

        fpr = np.zeros_like(fps, dtype=np.float64)
        np.divide(
            fps,
            negative_count[label_idx],
            where=negative_count[label_idx] > 0,
            out=fpr,
        )
        tpr = np.zeros_like(tps, dtype=np.float64)
        np.divide(
            tps,
            positive_count[label_idx],
            where=positive_count[label_idx] > 0,
            out=tpr,
        )

        # compute rocauc
        rocauc[label_idx] += npc.trapezoid(x=fpr, y=tpr, axis=0)

    return rocauc, prev


def compute_counts(
    ids: NDArray[np.int64],
    scores: NDArray[np.float64],
    winners: NDArray[np.bool_],
    score_thresholds: NDArray[np.float64],
    hardmax: bool,
    n_labels: int,
) -> NDArray[np.uint64]:
    """
    Computes counts of TP, FP and FN's per label.

    Parameters
    ----------
    ids : NDArray[np.int64]
        A sorted array of classification pairs with shape (n_pairs, 3).
            Index 0 - Datum Index
            Index 1 - GroundTruth Label Index
            Index 2 - Prediction Label Index
    scores : NDArray[np.float64]
        A sorted array of classification scores with shape (n_pairs,).
    winner : NDArray[np.bool_]
        Marks predictions with highest score over a datum.
    score_thresholds : NDArray[np.float64]
        A 1-D array contains score thresholds to compute metrics over.
    hardmax : bool
        Option to only allow a single positive prediction.
    n_labels : int
        The total number of unqiue labels.

    Returns
    -------
    NDArray[np.int32]
        TP, FP, FN, TN counts.
    """
    n_scores = score_thresholds.shape[0]
    counts = np.zeros((n_scores, n_labels, 4), dtype=np.uint64)
    if ids.size == 0:
        return counts

    gt_labels = ids[:, 1]
    pd_labels = ids[:, 2]

    mask_matching_labels = np.isclose(gt_labels, pd_labels)
    mask_score_nonzero = ~np.isclose(scores, 0.0)
    mask_hardmax = winners > 0.5
    mask_valid_gts = gt_labels >= 0
    mask_valid_pds = pd_labels >= 0

    # calculate metrics at various score thresholds
    for score_idx in range(n_scores):
        mask_score_threshold = scores >= score_thresholds[score_idx]
        mask_score = mask_score_nonzero & mask_score_threshold

        if hardmax:
            mask_score &= mask_hardmax

        mask_tp = mask_matching_labels & mask_score
        mask_fp = ~mask_matching_labels & mask_score
        mask_fn = (mask_matching_labels & ~mask_score) | mask_fp
        mask_tn = ~mask_matching_labels & ~mask_score

        mask_fn &= mask_valid_gts
        mask_fp &= mask_valid_pds

        fn = np.unique(ids[mask_fn][:, [0, 1]].astype(int), axis=0)
        tn = np.unique(ids[mask_tn][:, [0, 2]].astype(int), axis=0)

        counts[score_idx, :, 0] = np.bincount(
            pd_labels[mask_tp], minlength=n_labels
        )
        counts[score_idx, :, 1] = np.bincount(
            pd_labels[mask_fp], minlength=n_labels
        )
        counts[score_idx, :, 2] = np.bincount(fn[:, 1], minlength=n_labels)
        counts[score_idx, :, 3] = np.bincount(tn[:, 1], minlength=n_labels)

    return counts


def compute_precision(counts: NDArray[np.uint64]) -> NDArray[np.float64]:
    """
    Compute precision metric using result of compute_counts.
    """
    n_scores, n_labels, _ = counts.shape
    precision = np.zeros((n_scores, n_labels), dtype=np.float64)
    np.divide(
        counts[:, :, 0],
        (counts[:, :, 0] + counts[:, :, 1]),
        where=(counts[:, :, 0] + counts[:, :, 1]) > 0,
        out=precision,
    )
    return precision


def compute_recall(counts: NDArray[np.uint64]) -> NDArray[np.float64]:
    """
    Compute recall metric using result of compute_counts.
    """
    n_scores, n_labels, _ = counts.shape
    recall = np.zeros((n_scores, n_labels), dtype=np.float64)
    np.divide(
        counts[:, :, 0],
        (counts[:, :, 0] + counts[:, :, 2]),
        where=(counts[:, :, 0] + counts[:, :, 2]) > 0,
        out=recall,
    )
    return recall


def compute_f1_score(
    precision: NDArray[np.float64], recall: NDArray[np.float64]
) -> NDArray[np.float64]:
    """
    Compute f1 metric using result of compute_precision and compute_recall.
    """
    f1_score = np.zeros_like(recall)
    np.divide(
        (2 * precision * recall),
        (precision + recall),
        where=(precision + recall) > 1e-9,
        out=f1_score,
    )
    return f1_score


def compute_accuracy(
    counts: NDArray[np.uint64], n_datums: int
) -> NDArray[np.float64]:
    """
    Compute accuracy metric using result of compute_counts.
    """
    n_scores, _, _ = counts.shape
    accuracy = np.zeros(n_scores, dtype=np.float64)
    if n_datums == 0:
        return accuracy
    np.divide(
        counts[:, :, 0].sum(axis=1),
        n_datums,
        out=accuracy,
    )
    return accuracy


class PairClassification(IntFlag):
    TP = 1 << 0
    FP_FN_MISCLF = 1 << 1
    FN_UNMATCHED = 1 << 2


def compute_pair_classifications(
    ids: NDArray[np.int64],
    scores: NDArray[np.float64],
    winners: NDArray[np.bool_],
    score_thresholds: NDArray[np.float64],
    hardmax: bool,
) -> tuple[NDArray[np.bool_], NDArray[np.bool_], NDArray[np.bool_]]:
    """
    Classifiy ID pairs as TP, FP or FN.

    Parameters
    ----------
    ids : NDArray[np.int64]
        A sorted array of classification pairs with shape (n_pairs, 3).
            Index 0 - Datum Index
            Index 1 - GroundTruth Label Index
            Index 2 - Prediction Label Index
    scores : NDArray[np.float64]
        A sorted array of classification scores with shape (n_pairs,).
    winner : NDArray[np.bool_]
        Marks predictions with highest score over a datum.
    score_thresholds : NDArray[np.float64]
        A 1-D array containing score thresholds.
    hardmax : bool
        Option to only allow a single positive prediction.

    Returns
    -------
    NDArray[bool]
        True-positive mask.
    NDArray[bool]
        Misclassification FP, FN mask.
    NDArray[bool]
        Unmatched FN mask.
    """
    n_pairs = ids.shape[0]
    n_scores = score_thresholds.shape[0]

    gt_labels = ids[:, 1]
    pd_labels = ids[:, 2]
    groundtruths = ids[:, [0, 1]]

    pair_classifications = np.zeros(
        (n_scores, n_pairs),
        dtype=np.uint8,
    )

    mask_label_match = np.isclose(gt_labels, pd_labels)
    mask_score = scores > 1e-9
    for score_idx in range(n_scores):
        mask_score &= scores >= score_thresholds[score_idx]
        if hardmax:
            mask_score &= winners

        mask_true_positives = mask_label_match & mask_score
        mask_misclassifications = ~mask_label_match & mask_score
        mask_unmatched_groundtruths = ~(
            (
                groundtruths.reshape(-1, 1, 2)
                == groundtruths[mask_score].reshape(1, -1, 2)
            )
            .all(axis=2)
            .any(axis=1)
        )

        # classify pairings
        pair_classifications[score_idx, mask_true_positives] |= np.uint8(
            PairClassification.TP
        )
        pair_classifications[score_idx, mask_misclassifications] |= np.uint8(
            PairClassification.FP_FN_MISCLF
        )
        pair_classifications[
            score_idx, mask_unmatched_groundtruths
        ] |= np.uint8(PairClassification.FN_UNMATCHED)

    mask_tp = np.bitwise_and(pair_classifications, PairClassification.TP) > 0
    mask_fp_fn_misclf = (
        np.bitwise_and(pair_classifications, PairClassification.FP_FN_MISCLF)
        > 0
    )
    mask_fn_unmatched = (
        np.bitwise_and(pair_classifications, PairClassification.FN_UNMATCHED)
        > 0
    )

    return (
        mask_tp,
        mask_fp_fn_misclf,
        mask_fn_unmatched,
    )


def compute_confusion_matrix(
    ids: NDArray[np.int64],
    mask_tp: NDArray[np.bool_],
    mask_fp_fn_misclf: NDArray[np.bool_],
    mask_fn_unmatched: NDArray[np.bool_],
    score_thresholds: NDArray[np.float64],
    n_labels: int,
):
    """
    Compute confusion matrix using output of compute_pair_classifications.
    """
    n_scores = score_thresholds.size

    # initialize arrays
    confusion_matrices = np.zeros(
        (n_scores, n_labels, n_labels), dtype=np.uint64
    )
    unmatched_groundtruths = np.zeros((n_scores, n_labels), dtype=np.uint64)

    mask_matched = mask_tp | mask_fp_fn_misclf
    for score_idx in range(n_scores):
        # matched annotations
        unique_pairs = np.unique(
            ids[np.ix_(mask_matched[score_idx], (0, 1, 2))],  # type: ignore - numpy ix_ typing
            axis=0,
        )
        unique_labels, unique_label_counts = np.unique(
            unique_pairs[:, (1, 2)], axis=0, return_counts=True
        )
        confusion_matrices[
            score_idx, unique_labels[:, 0], unique_labels[:, 1]
        ] = unique_label_counts

        # unmatched groundtruths
        unique_pairs = np.unique(
            ids[np.ix_(mask_fn_unmatched[score_idx], (0, 1))],  # type: ignore - numpy ix_ typing
            axis=0,
        )
        unique_labels, unique_label_counts = np.unique(
            unique_pairs[:, 1], return_counts=True
        )
        unmatched_groundtruths[score_idx, unique_labels] = unique_label_counts

    return confusion_matrices, unmatched_groundtruths



from functools import wraps

from valor_lite.text_generation.annotation import QueryResponse
from valor_lite.text_generation.computation import (
    calculate_answer_correctness,
    calculate_answer_relevance,
    calculate_bias,
    calculate_context_precision,
    calculate_context_recall,
    calculate_context_relevance,
    calculate_faithfulness,
    calculate_hallucination,
    calculate_rouge_scores,
    calculate_sentence_bleu,
    calculate_summary_coherence,
    calculate_toxicity,
)
from valor_lite.text_generation.llm.exceptions import InvalidLLMResponseError
from valor_lite.text_generation.llm.integrations import (
    ClientWrapper,
    MistralWrapper,
    OpenAIWrapper,
)
from valor_lite.text_generation.metric import Metric, MetricType


def llm_guided_metric(fn):
    """
    Call the LLMClient class function with retries for InvalidLLMResponseError.

    If retries is set to 0, then the function will only be called once and not retried.

    If, for example, retries is set to 3, then the function will be retried in the
    event of an InvalidLLMResponseError up to 3 times, for a maximum of 4 calls.
    """

    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        client = getattr(self, "client", None)
        if client is None:
            raise ValueError(
                f"{fn.__name__} requires the definition of an LLM client."
            )
        if getattr(client, "model_name", None) is None:
            raise AttributeError(
                "Client wrapper should contain 'model_name' as a string attribute."
            )

        error = None
        retries = getattr(self, "retries", 0)
        for _ in range(1 + retries):
            try:
                return fn(self, *args, **kwargs)
            except InvalidLLMResponseError as e:
                error = e
        if error is not None:
            return Metric.error(
                error_type=type(error).__name__,
                error_message=str(error),
                model_name=client.model_name,
                retries=retries,
            )

    return wrapper


class Evaluator:
    """
    Parent class for all LLM clients.

    Attributes
    ----------
    client : ClientWrapper, optional
        An optional client to compute llm-guided metrics.
    retries : int
        The number of times to retry the API call if it fails. Defaults to 0, indicating
        that the call will not be retried.
    """

    def __init__(
        self,
        client: ClientWrapper | None = None,
        retries: int = 0,
        default_system_prompt: str = "You are a helpful assistant.",
    ):
        """
        Creates an instance of a generic LLM client.

        Parameters
        ----------
        client : ClientWrapper, optional
            Any LLM client that conforms to _ClientWrapper. Required for LLM-guided metrics.
        retries : int, default=0
            The number of times to retry the API call if it fails. Defaults to 0, indicating
            that the call will not be retried.
        default_system_prompt : str, default="You are a helpful assistant."
            The default system prompt that is given to the evaluating LLM.
        """

        self.client = client
        self.retries = retries
        self.default_system_prompt = default_system_prompt

    @classmethod
    def openai(
        cls,
        model_name: str = "gpt-3.5-turbo",
        api_key: str | None = None,
        retries: int = 0,
        seed: int | None = None,
        default_system_prompt: str = "You are a helpful assistant.",
    ):
        """
        Create an evaluator using OpenAI's client.

        Parameters
        ----------
        model_name : str, default="gpt-3.5-turbo"
            The model to use. Defaults to "gpt-3.5-turbo".
        api_key : str, optional
            The OpenAI API key to use. If not specified, then the OPENAI_API_KEY environment
            variable will be used.
        retries : int, default=0
            The number of times to retry the API call if it fails. Defaults to 0, indicating
            that the call will not be retried. For example, if self.retries is set to 3,
            this means that the call will be retried up to 3 times, for a maximum of 4 calls.
        seed : int, optional
            An optional seed can be provided to GPT to get deterministic results.
        default_system_prompt : str, default="You are a helpful assistant."
            The default system prompt that is given to the evaluating LLM.
        """
        if seed is not None:
            if retries != 0:
                raise ValueError(
                    "Seed is provided, but retries is not 0. Retries should be 0 when seed is provided."
                )
        client = OpenAIWrapper(
            api_key=api_key,
            model_name=model_name,
            seed=seed,
        )
        return cls(
            client=client,
            retries=retries,
            default_system_prompt=default_system_prompt,
        )

    @classmethod
    def mistral(
        cls,
        model_name: str = "mistral-small-latest",
        api_key: str | None = None,
        retries: int = 0,
        default_system_prompt: str = "You are a helpful assistant.",
    ):
        """
        Create an evaluator using the Mistral API.

        Parameters
        ----------
        model_name : str, default="mistral-small-latest"
            The model to use. Defaults to "mistral-small-latest".
        api_key : str, optional
            The Mistral API key to use. If not specified, then the MISTRAL_API_KEY environment
            variable will be used.
        retries : int, default=0
            The number of times to retry the API call if it fails. Defaults to 0, indicating
            that the call will not be retried. For example, if self.retries is set to 3,
            this means that the call will be retried up to 3 times, for a maximum of 4 calls.
        default_system_prompt : str, default="You are a helpful assistant."
            The default system prompt that is given to the evaluating LLM.
        """
        client = MistralWrapper(
            api_key=api_key,
            model_name=model_name,
        )
        return cls(
            client=client,
            retries=retries,
            default_system_prompt=default_system_prompt,
        )

    @llm_guided_metric
    def compute_answer_correctness(
        self,
        response: QueryResponse,
    ) -> Metric:
        """
        Compute answer correctness. Answer correctness is computed as an f1 score obtained
        by comparing prediction statements to ground truth statements.

        If there are multiple ground truths, then the f1 score is computed for each ground
        truth and the maximum score is returned.

        This metric was adapted from RAGAS. We follow a similar prompting strategy and
        computation, however we do not do a weighted sum with an answer similarity score
        using embeddings.

        Parameters
        ----------
        response: QueryResponse
            A user query with ground truth and generated response.

        Returns
        -------
        Metric
            The answer correctness score between 0 and 1. Higher values indicate that the
            answer is more correct. A score of 1 indicates that all statements in the
            prediction are supported by the ground truth and all statements in the ground
            truth are present in the prediction.
        """
        if not response.context:
            raise ValueError("The answer correctness metric requires context.")

        result = calculate_answer_correctness(
            client=self.client,  # type: ignore - wrapper handles None case
            system_prompt=self.default_system_prompt,
            query=response.query,
            response=response.response,
            groundtruths=response.context.groundtruth,
        )
        return Metric.answer_correctness(
            value=result,
            model_name=self.client.model_name,  # type: ignore - wrapper handles None case
            retries=self.retries,
        )

    @llm_guided_metric
    def compute_answer_relevance(self, response: QueryResponse) -> Metric:
        """
        Compute answer relevance, the proportion of the model response that is
        relevant to the query, for a single piece of text.

        Parameters
        ----------
        response: QueryResponse
            A user query with ground truth and generated response.

        Returns
        -------
        Metric
            The answer relevance score between 0 and 1. A score of 1 indicates that all
            statements are relevant to the query.
        """
        result = calculate_answer_relevance(
            client=self.client,  # type: ignore - wrapper handles None case
            system_prompt=self.default_system_prompt,
            query=response.query,
            response=response.response,
        )
        return Metric.answer_relevance(
            value=result,
            model_name=self.client.model_name,  # type: ignore - wrapper handles None case
            retries=self.retries,
        )

    @llm_guided_metric
    def compute_bias(
        self,
        response: QueryResponse,
    ) -> Metric:
        """
        Compute bias, the proportion of model opinions that are biased.

        Parameters
        ----------
        response: QueryResponse
            A user query with ground truth and generated response.

        Returns
        -------
        float
            The bias score between 0 and 1. A score of 1 indicates that all opinions in
            the text are biased.
        """
        result = calculate_bias(
            client=self.client,  # type: ignore - wrapper handles None case
            system_prompt=self.default_system_prompt,
            response=response.response,
        )
        return Metric.bias(
            value=result,
            model_name=self.client.model_name,  # type: ignore - wrapper handles None case
            retries=self.retries,
        )

    @llm_guided_metric
    def compute_context_precision(
        self,
        response: QueryResponse,
    ) -> Metric:
        """
        Compute context precision, a score for evaluating the retrieval
        mechanism of a RAG model.

        First, an LLM is prompted to determine if each context in the context
        list is useful for producing the ground truth answer to the query.

        If there are multiple ground truths, then the verdict is "yes" for a
        context if that context is useful for producing any of the ground truth
        answers, and "no" otherwise.

        Then, using these verdicts, the context precision score is computed as
        a weighted sum of the precision at k for each k from 1 to the length
        of the context list.

        Note that the earlier a piece of context appears in the context list,
        the more important it is in the computation of this score. For example,
        the first context in the context list will be included in every precision
        at k computation, so will have a large influence on the final score,
        whereas the last context will only be used for the last precision at
        k computation, so will have a small influence on the final score.

        Parameters
        ----------
        response: QueryResponse
            A user query with ground truth and generated response.

        Returns
        -------
        Metric
            The context precision score between 0 and 1. A higher score indicates
            better context precision.
        """
        if not response.context:
            raise ValueError("The context precision metric requires context.")

        result = calculate_context_precision(
            client=self.client,  # type: ignore - wrapper handles None case
            system_prompt=self.default_system_prompt,
            query=response.query,
            predicted_context=response.context.prediction,
            groundtruth_context=response.context.groundtruth,
        )
        return Metric.context_precision(
            value=result,
            model_name=self.client.model_name,  # type: ignore - wrapper handles None case
            retries=self.retries,
        )

    @llm_guided_metric
    def compute_context_recall(
        self,
        response: QueryResponse,
    ) -> Metric:
        """
        Compute context recall, a score for evaluating the retrieval mechanism of a RAG model.

        The context recall score is the proportion of statements in the ground truth
        that are attributable to the context list.

        If multiple ground truths are provided, then the context recall score is
        computed for each ground truth and the maximum score is returned.

        Parameters
        ----------
        response: QueryResponse
            A user query with ground truth and generated response.

        Returns
        -------
        Metric
            The context recall score between 0 and 1. A score of 1 indicates that
            all ground truth statements are attributable to the contexts in the context list.
        """
        if not response.context:
            raise ValueError("The context recall metric requires context.")

        result = calculate_context_recall(
            client=self.client,  # type: ignore - wrapper handles None case
            system_prompt=self.default_system_prompt,
            predicted_context=response.context.prediction,
            groundtruth_context=response.context.groundtruth,
        )
        return Metric.context_recall(
            value=result,
            model_name=self.client.model_name,  # type: ignore - wrapper handles None case
            retries=self.retries,
        )

    @llm_guided_metric
    def compute_context_relevance(
        self,
        response: QueryResponse,
    ) -> Metric:
        """
        Compute context relevance, the proportion of contexts in the context list
        that are relevant to the query.

        Parameters
        ----------
        response: QueryResponse
            A user query with ground truth and generated response.

        Returns
        -------
        Metric
            The context relevance score between 0 and 1. A score of 0 indicates
            that none of the contexts are relevant and a score of 1 indicates
            that all of the contexts are relevant.
        """
        if not response.context:
            raise ValueError("The context relevance metric requires context.")

        result = calculate_context_relevance(
            client=self.client,  # type: ignore - wrapper handles None case
            system_prompt=self.default_system_prompt,
            query=response.query,
            context=response.context.prediction,
        )
        return Metric.context_relevance(
            value=result,
            model_name=self.client.model_name,  # type: ignore - wrapper handles None case
            retries=self.retries,
        )

    @llm_guided_metric
    def compute_faithfulness(
        self,
        response: QueryResponse,
    ) -> Metric:
        """
        Compute the faithfulness score. The faithfulness score is the proportion
        of claims in the text that are implied by the list of contexts. Claims
        that contradict the list of contexts and claims that are unrelated to
        the list of contexts both count against the score.

        Parameters
        ----------
        response: QueryResponse
            A user query with ground truth and generated response.

        Returns
        -------
        Metric
            The faithfulness score between 0 and 1. A score of 1 indicates that
            all claims in the text are implied by the list of contexts.
        """

        if not response.context:
            raise ValueError("The faithfulness metric requires context.")

        result = calculate_faithfulness(
            client=self.client,  # type: ignore - wrapper handles None case
            system_prompt=self.default_system_prompt,
            response=response.response,
            context=response.context.prediction,
        )
        return Metric.faithfulness(
            value=result,
            model_name=self.client.model_name,  # type: ignore - wrapper handles None case
            retries=self.retries,
        )

    @llm_guided_metric
    def compute_hallucination(
        self,
        response: QueryResponse,
    ) -> Metric:
        """
        Compute the hallucination score, the proportion of contexts in the context
        list that are contradicted by the text.

        Parameters
        ----------
        response: QueryResponse
            A user query with ground truth and generated response.

        Returns
        -------
        Metric
            The hallucination score between 0 and 1. A score of 1 indicates that
            all contexts are contradicted by the text.
        """

        if not response.context:
            raise ValueError("The hallucination metric requires context.")

        result = calculate_hallucination(
            client=self.client,  # type: ignore - wrapper handles None case
            system_prompt=self.default_system_prompt,
            response=response.response,
            context=response.context.prediction,
        )
        return Metric.hallucination(
            value=result,
            model_name=self.client.model_name,  # type: ignore - wrapper handles None case
            retries=self.retries,
        )

    @llm_guided_metric
    def compute_summary_coherence(
        self,
        response: QueryResponse,
    ) -> Metric:
        """
        Compute summary coherence, the collective quality of a summary.

        Parameters
        ----------
        response: QueryResponse
            A user query with ground truth and generated response.

        Returns
        -------
        Metric
            The summary coherence score between 1 and 5. A score of 1 indicates
            the lowest summary coherence and a score of 5 indicates the highest
            summary coherence.
        """
        result = calculate_summary_coherence(
            client=self.client,  # type: ignore - wrapper handles None case
            system_prompt=self.default_system_prompt,
            text=response.query,
            summary=response.response,
        )
        return Metric.summary_coherence(
            value=result,
            model_name=self.client.model_name,  # type: ignore - wrapper handles None case
            retries=self.retries,
        )

    @llm_guided_metric
    def compute_toxicity(
        self,
        response: QueryResponse,
    ) -> Metric:
        """
        Compute toxicity, the portion of opinions that are toxic.

        Parameters
        ----------
        response: QueryResponse
            A user query with ground truth and generated response.

        Returns
        -------
        Metric
            The toxicity score will be evaluated as a float between 0 and 1, with
            1 indicating that all opinions in the text are toxic.
        """
        result = calculate_toxicity(
            client=self.client,  # type: ignore - wrapper handles None case
            system_prompt=self.default_system_prompt,
            response=response.response,
        )
        return Metric.toxicity(
            value=result,
            model_name=self.client.model_name,  # type: ignore - wrapper handles None case
            retries=self.retries,
        )

    @staticmethod
    def compute_rouge(
        response: QueryResponse,
        rouge_types: list[str] = [
            "rouge1",
            "rouge2",
            "rougeL",
            "rougeLsum",
        ],
        use_stemmer: bool = False,
    ) -> list[Metric]:
        """
        Calculate ROUGE scores for a model response given some set of references.

        Parameters
        ----------
        response: QueryResponse
            A user query with ground truth and generated response.
        rouge_types : list[str], optional
            A list of rouge types to calculate.
            Defaults to ['rouge1', 'rouge2', 'rougeL', 'rougeLsum'].
        use_stemmer: bool, default=False
            If True, uses Porter stemmer to strip word suffixes. Defaults to False.

        Returns
        -------
        list[Metric]
        """

        if not response.context:
            raise ValueError("ROUGE metrics require context.")

        results = calculate_rouge_scores(
            prediction=response.response,
            references=response.context.groundtruth,
            rouge_types=rouge_types,
            use_stemmer=use_stemmer,
        )
        return [
            Metric.rouge(
                value=result,
                rouge_type=rouge_type,
                use_stemmer=use_stemmer,
            )
            for rouge_type, result in results.items()
        ]

    @staticmethod
    def compute_sentence_bleu(
        response: QueryResponse,
        weights: list[float] = [0.25, 0.25, 0.25, 0.25],
    ) -> Metric:
        """
        Calculate sentence BLEU scores for a set of model response - ground truth pairs.

        Parameters
        ----------
        response: QueryResponse
            A user query with ground truth and generated response.
        weights: list[float], default=[0.25, 0.25, 0.25, 0.25]
            The default BLEU calculates a score for up to 4-grams using uniform
            weights (this is called BLEU-4). To evaluate your translations with
            higher/lower order ngrams, use customized weights. Example: when accounting
            for up to 5-grams with uniform weights (this is called BLEU-5) use [1/5]*5
        """

        if not response.context:
            raise ValueError("The sentence BLEU metric requires context.")

        result = calculate_sentence_bleu(
            prediction=response.response,
            references=response.context.groundtruth,
            weights=weights,
        )
        return Metric.bleu(
            value=result,
            weights=weights,
        )

    def compute_all(
        self,
        response: QueryResponse,
        bleu_weights: list[float] = [0.25, 0.25, 0.25, 0.25],
        rouge_types: list[str] = [
            "rouge1",
            "rouge2",
            "rougeL",
            "rougeLsum",
        ],
        rouge_use_stemmer: bool = False,
    ) -> dict[MetricType, list[Metric]]:
        """
        Computes all available metrics.

        Parameters
        ----------
        response: QueryResponse
            A user query with ground truth and generated response.
        bleu_weights: list[float], default=[0.25, 0.25, 0.25, 0.25]
            The default BLEU calculates a score for up to 4-grams using uniform
            weights (this is called BLEU-4). To evaluate your translations with
            higher/lower order ngrams, use customized weights. Example: when accounting
            for up to 5-grams with uniform weights (this is called BLEU-5) use [1/5]*5
        rouge_types : list[str], optional
            A list of rouge types to calculate.
            Defaults to ['rouge1', 'rouge2', 'rougeL', 'rougeLsum'].
        rouge_use_stemmer: bool, default=False
            If True, uses Porter stemmer to strip word suffixes. Defaults to False.
        """
        results = dict()
        results[MetricType.AnswerCorrectness] = [
            self.compute_answer_correctness(response)
        ]
        results[MetricType.AnswerRelevance] = [
            self.compute_answer_relevance(response)
        ]
        results[MetricType.Bias] = [self.compute_bias(response)]
        results[MetricType.ContextPrecision] = [
            self.compute_context_precision(response)
        ]
        results[MetricType.ContextRecall] = [
            self.compute_context_recall(response)
        ]
        results[MetricType.ContextRelevance] = [
            self.compute_context_relevance(response)
        ]
        results[MetricType.Faithfulness] = [
            self.compute_faithfulness(response)
        ]
        results[MetricType.Hallucination] = [
            self.compute_hallucination(response)
        ]
        results[MetricType.SummaryCoherence] = [
            self.compute_summary_coherence(response)
        ]
        results[MetricType.Toxicity] = [self.compute_toxicity(response)]
        results[MetricType.ROUGE] = self.compute_rouge(
            response=response,
            rouge_types=rouge_types,
            use_stemmer=rouge_use_stemmer,
        )
        results[MetricType.BLEU] = [
            self.compute_sentence_bleu(response=response, weights=bleu_weights)
        ]
        return results



import numpy as np
from numpy.typing import NDArray

try:
    _numpy_trapezoid = np.trapezoid  # numpy v2
except AttributeError:
    _numpy_trapezoid = np.trapz  # numpy v1


def trapezoid(
    x: NDArray[np.float64], y: NDArray[np.float64], axis: int
) -> NDArray[np.float64]:
    return _numpy_trapezoid(x=x, y=y, axis=axis)  # type: ignore - NumPy compatibility



from .annotation import Classification
from .evaluator import Evaluator
from .loader import Loader
from .metric import Metric, MetricType
from .shared import EvaluatorInfo

__all__ = [
    "Classification",
    "MetricType",
    "Loader",
    "Evaluator",
    "Metric",
    "EvaluatorInfo",
]



from dataclasses import dataclass
from enum import Enum

from valor_lite.schemas import BaseMetric


class MetricType(str, Enum):
    Counts = "Counts"
    Precision = "Precision"
    Recall = "Recall"
    F1 = "F1"
    AP = "AP"
    AR = "AR"
    mAP = "mAP"
    mAR = "mAR"
    APAveragedOverIOUs = "APAveragedOverIOUs"
    mAPAveragedOverIOUs = "mAPAveragedOverIOUs"
    ARAveragedOverScores = "ARAveragedOverScores"
    mARAveragedOverScores = "mARAveragedOverScores"
    PrecisionRecallCurve = "PrecisionRecallCurve"
    ConfusionMatrixWithExamples = "ConfusionMatrixWithExamples"
    ConfusionMatrix = "ConfusionMatrix"
    Examples = "Examples"


@dataclass
class Metric(BaseMetric):
    """
    Object Detection Metric.

    Attributes
    ----------
    type : str
        The metric type.
    value : int | float | dict
        The metric value.
    parameters : dict[str, Any]
        A dictionary containing metric parameters.
    """

    def __post_init__(self):
        if not isinstance(self.type, str):
            raise TypeError(
                f"Metric type should be of type 'str': {self.type}"
            )
        elif not isinstance(self.value, (int, float, dict)):
            raise TypeError(
                f"Metric value must be of type 'int', 'float' or 'dict': {self.value}"
            )
        elif not isinstance(self.parameters, dict):
            raise TypeError(
                f"Metric parameters must be of type 'dict[str, Any]': {self.parameters}"
            )
        elif not all([isinstance(k, str) for k in self.parameters.keys()]):
            raise TypeError(
                f"Metric parameter dictionary should only have keys with type 'str': {self.parameters}"
            )

    @classmethod
    def precision(
        cls,
        value: float,
        label: str,
        iou_threshold: float,
        score_threshold: float,
    ):
        """
        Precision metric for a specific class label in object detection.

        This class encapsulates a metric value for a particular class label,
        along with the associated Intersection over Union (IOU) threshold and
        confidence score threshold.

        Parameters
        ----------
        value : float
            The metric value.
        label : str
            The class label for which the metric is calculated.
        iou_threshold : float
            The IOU threshold used to determine matches between predicted and ground truth boxes.
        score_threshold : float
            The confidence score threshold above which predictions are considered.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.Precision.value,
            value=value,
            parameters={
                "label": label,
                "iou_threshold": iou_threshold,
                "score_threshold": score_threshold,
            },
        )

    @classmethod
    def recall(
        cls,
        value: float,
        label: str,
        iou_threshold: float,
        score_threshold: float,
    ):
        """
        Recall metric for a specific class label in object detection.

        This class encapsulates a metric value for a particular class label,
        along with the associated Intersection over Union (IOU) threshold and
        confidence score threshold.

        Parameters
        ----------
        value : float
            The metric value.
        label : str
            The class label for which the metric is calculated.
        iou_threshold : float
            The IOU threshold used to determine matches between predicted and ground truth boxes.
        score_threshold : float
            The confidence score threshold above which predictions are considered.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.Recall.value,
            value=value,
            parameters={
                "label": label,
                "iou_threshold": iou_threshold,
                "score_threshold": score_threshold,
            },
        )

    @classmethod
    def f1_score(
        cls,
        value: float,
        label: str,
        iou_threshold: float,
        score_threshold: float,
    ):
        """
        F1 score for a specific class label in object detection.

        This class encapsulates a metric value for a particular class label,
        along with the associated Intersection over Union (IOU) threshold and
        confidence score threshold.

        Parameters
        ----------
        value : float
            The metric value.
        label : str
            The class label for which the metric is calculated.
        iou_threshold : float
            The IOU threshold used to determine matches between predicted and ground truth boxes.
        score_threshold : float
            The confidence score threshold above which predictions are considered.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.F1.value,
            value=value,
            parameters={
                "label": label,
                "iou_threshold": iou_threshold,
                "score_threshold": score_threshold,
            },
        )

    @classmethod
    def average_precision(
        cls,
        value: float,
        iou_threshold: float,
        label: str,
    ):
        """
        Average Precision (AP) metric for object detection tasks.

        The AP computation uses 101-point interpolation, which calculates the average
        precision by interpolating the precision-recall curve at 101 evenly spaced recall
        levels from 0 to 1.

        Parameters
        ----------
        value : float
            The average precision value.
        iou_threshold : float
            The IOU threshold used to compute the AP.
        label : str
            The class label for which the AP is computed.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.AP.value,
            value=value,
            parameters={
                "iou_threshold": iou_threshold,
                "label": label,
            },
        )

    @classmethod
    def mean_average_precision(
        cls,
        value: float,
        iou_threshold: float,
    ):
        """
        Mean Average Precision (mAP) metric for object detection tasks.

        The AP computation uses 101-point interpolation, which calculates the average
        precision for each class by interpolating the precision-recall curve at 101 evenly
        spaced recall levels from 0 to 1. The mAP is then calculated by averaging these
        values across all class labels.

        Parameters
        ----------
        value : float
            The mean average precision value.
        iou_threshold : float
            The IOU threshold used to compute the mAP.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.mAP.value,
            value=value,
            parameters={
                "iou_threshold": iou_threshold,
            },
        )

    @classmethod
    def average_precision_averaged_over_IOUs(
        cls,
        value: float,
        iou_thresholds: list[float],
        label: str,
    ):
        """
        Average Precision (AP) metric averaged over multiple IOU thresholds.

        The AP computation uses 101-point interpolation, which calculates the average precision
        by interpolating the precision-recall curve at 101 evenly spaced recall levels from 0 to 1
        for each IOU threshold specified in `iou_thresholds`. The final APAveragedOverIOUs value is
        obtained by averaging these AP values across all specified IOU thresholds.

        Parameters
        ----------
        value : float
            The average precision value averaged over the specified IOU thresholds.
        iou_thresholds : list[float]
            The list of IOU thresholds used to compute the AP values.
        label : str
            The class label for which the AP is computed.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.APAveragedOverIOUs.value,
            value=value,
            parameters={
                "iou_thresholds": iou_thresholds,
                "label": label,
            },
        )

    @classmethod
    def mean_average_precision_averaged_over_IOUs(
        cls,
        value: float,
        iou_thresholds: list[float],
    ):
        """
        Mean Average Precision (mAP) metric averaged over multiple IOU thresholds.

        The AP computation uses 101-point interpolation, which calculates the average precision
        by interpolating the precision-recall curve at 101 evenly spaced recall levels from 0 to 1
        for each IOU threshold specified in `iou_thresholds`. The final mAPAveragedOverIOUs value is
        obtained by averaging these AP values across all specified IOU thresholds and all class labels.

        Parameters
        ----------
        value : float
            The average precision value averaged over the specified IOU thresholds.
        iou_thresholds : list[float]
            The list of IOU thresholds used to compute the AP values.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.mAPAveragedOverIOUs.value,
            value=value,
            parameters={
                "iou_thresholds": iou_thresholds,
            },
        )

    @classmethod
    def average_recall(
        cls,
        value: float,
        score_threshold: float,
        iou_thresholds: list[float],
        label: str,
    ):
        """
        Average Recall (AR) metric for object detection tasks.

        The AR computation considers detections with confidence scores above the specified
        `score_threshold` and calculates the recall at each IOU threshold in `iou_thresholds`.
        The final AR value is the average of these recall values across all specified IOU
        thresholds.

        Parameters
        ----------
        value : float
            The average recall value averaged over the specified IOU thresholds.
        score_threshold : float
            The detection score threshold; only detections with confidence scores above this
            threshold are considered.
        iou_thresholds : list[float]
            The list of IOU thresholds used to compute the recall values.
        label : str
            The class label for which the AR is computed.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.AR.value,
            value=value,
            parameters={
                "iou_thresholds": iou_thresholds,
                "score_threshold": score_threshold,
                "label": label,
            },
        )

    @classmethod
    def mean_average_recall(
        cls,
        value: float,
        score_threshold: float,
        iou_thresholds: list[float],
    ):
        """
        Mean Average Recall (mAR) metric for object detection tasks.

        The mAR computation considers detections with confidence scores above the specified
        `score_threshold` and calculates recall at each IOU threshold in `iou_thresholds` for
        each label. The final mAR value is obtained by averaging these recall values over the
        specified IOU thresholds and then averaging across all labels.

        Parameters
        ----------
        value : float
            The mean average recall value averaged over the specified IOU thresholds.
        score_threshold : float
            The detection score threshold; only detections with confidence scores above this
            threshold are considered.
        iou_thresholds : list[float]
            The list of IOU thresholds used to compute the recall values.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.mAR.value,
            value=value,
            parameters={
                "iou_thresholds": iou_thresholds,
                "score_threshold": score_threshold,
            },
        )

    @classmethod
    def average_recall_averaged_over_scores(
        cls,
        value: float,
        score_thresholds: list[float],
        iou_thresholds: list[float],
        label: str,
    ):
        """
        Average Recall (AR) metric averaged over multiple score thresholds for a specific object class label.

        The AR computation considers detections across multiple `score_thresholds` and calculates
        recall at each IOU threshold in `iou_thresholds`. The final AR value is obtained by averaging
        the recall values over all specified score thresholds and IOU thresholds.

        Parameters
        ----------
        value : float
            The average recall value averaged over the specified score thresholds and IOU thresholds.
        score_thresholds : list[float]
            The list of detection score thresholds; detections with confidence scores above each threshold are considered.
        iou_thresholds : list[float]
            The list of IOU thresholds used to compute the recall values.
        label : str
            The class label for which the AR is computed.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.ARAveragedOverScores.value,
            value=value,
            parameters={
                "iou_thresholds": iou_thresholds,
                "score_thresholds": score_thresholds,
                "label": label,
            },
        )

    @classmethod
    def mean_average_recall_averaged_over_scores(
        cls,
        value: float,
        score_thresholds: list[float],
        iou_thresholds: list[float],
    ):
        """
        Mean Average Recall (mAR) metric averaged over multiple score thresholds and IOU thresholds.

        The mAR computation considers detections across multiple `score_thresholds`, calculates recall
        at each IOU threshold in `iou_thresholds` for each label, averages these recall values over all
        specified score thresholds and IOU thresholds, and then computes the mean across all labels to
        obtain the final mAR value.

        Parameters
        ----------
        value : float
            The mean average recall value averaged over the specified score thresholds and IOU thresholds.
        score_thresholds : list[float]
            The list of detection score thresholds; detections with confidence scores above each threshold are considered.
        iou_thresholds : list[float]
            The list of IOU thresholds used to compute the recall values.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.mARAveragedOverScores.value,
            value=value,
            parameters={
                "iou_thresholds": iou_thresholds,
                "score_thresholds": score_thresholds,
            },
        )

    @classmethod
    def precision_recall_curve(
        cls,
        precisions: list[float],
        scores: list[float],
        iou_threshold: float,
        label: str,
    ):
        """
        Interpolated precision-recall curve over 101 recall points.

        The precision values are interpolated over recalls ranging from 0.0 to 1.0 in steps of 0.01,
        resulting in 101 points. This is a byproduct of the 101-point interpolation used in calculating
        the Average Precision (AP) metric in object detection tasks.

        Parameters
        ----------
        precisions : list[float]
            Interpolated precision values corresponding to recalls at 0.0, 0.01, ..., 1.0.
        scores : list[float]
            Maximum prediction score for each point on the interpolated curve.
        iou_threshold : float
            The Intersection over Union (IOU) threshold used to determine true positives.
        label : str
            The class label associated with this precision-recall curve.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.PrecisionRecallCurve.value,
            value={
                "precisions": precisions,
                "scores": scores,
            },
            parameters={
                "iou_threshold": iou_threshold,
                "label": label,
            },
        )

    @classmethod
    def counts(
        cls,
        tp: int,
        fp: int,
        fn: int,
        label: str,
        iou_threshold: float,
        score_threshold: float,
    ):
        """
        `Counts` encapsulates the counts of true positives (`tp`), false positives (`fp`),
        and false negatives (`fn`) for object detection evaluation, along with the associated
        class label, Intersection over Union (IOU) threshold, and confidence score threshold.

        Parameters
        ----------
        tp : int
            Number of true positives.
        fp : int
            Number of false positives.
        fn : int
            Number of false negatives.
        label : str
            The class label for which the counts are calculated.
        iou_threshold : float
            The IOU threshold used to determine a match between predicted and ground truth boxes.
        score_threshold : float
            The confidence score threshold above which predictions are considered.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.Counts.value,
            value={
                "tp": tp,
                "fp": fp,
                "fn": fn,
            },
            parameters={
                "iou_threshold": iou_threshold,
                "score_threshold": score_threshold,
                "label": label,
            },
        )

    @classmethod
    def confusion_matrix(
        cls,
        confusion_matrix: dict[str, dict[str, int]],
        unmatched_predictions: dict[str, int],
        unmatched_ground_truths: dict[str, int],
        score_threshold: float,
        iou_threshold: float,
    ):
        """
        Confusion matrix for object detection task.

        This class encapsulates detailed information about the model's performance, including correct
        predictions, misclassifications, unmatched_predictions (subset of false positives), and unmatched ground truths
        (subset of false negatives).

        Confusion Matrix Format:
        {
            <ground truth label>: {
                <prediction label>: 129
                ...
            },
            ...
        }

        Unmatched Predictions Format:
        {
            <prediction label>: 11
            ...
        }

        Unmatched Ground Truths Format:
        {
            <ground truth label>: 7
            ...
        }

        Parameters
        ----------
        confusion_matrix : dict
            A nested dictionary containing integer counts of occurences where the first key is the ground truth label value
            and the second key is the prediction label value.
        unmatched_predictions : dict
            A dictionary where each key is a prediction label value with no corresponding ground truth
            (subset of false positives). The value is a dictionary containing counts.
        unmatched_ground_truths : dict
            A dictionary where each key is a ground truth label value for which the model failed to predict
            (subset of false negatives). The value is a dictionary containing counts.
        score_threshold : float
            The confidence score threshold used to filter predictions.
        iou_threshold : float
            The Intersection over Union (IOU) threshold used to determine true positives.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.ConfusionMatrix.value,
            value={
                "confusion_matrix": confusion_matrix,
                "unmatched_predictions": unmatched_predictions,
                "unmatched_ground_truths": unmatched_ground_truths,
            },
            parameters={
                "score_threshold": score_threshold,
                "iou_threshold": iou_threshold,
            },
        )

    @classmethod
    def examples(
        cls,
        datum_id: str,
        true_positives: list[tuple[str, str]],
        false_positives: list[str],
        false_negatives: list[str],
        score_threshold: float,
        iou_threshold: float,
    ):
        """
        Per-datum examples for object detection tasks.

        This metric is per-datum and contains lists of annotation identifiers that categorize them
        as true-positive, false-positive or false-negative. This is intended to be used with an
        external database where the identifiers can be used for retrieval.

        Examples Format:
        {
            "type": "Examples",
            "value": {
                "datum_id": "some string ID",
                "true_positives": [
                    ["groundtruth0", "prediction0"],
                    ["groundtruth123", "prediction11"],
                    ...
                ],
                "false_positives": [
                    "prediction25",
                    "prediction92",
                    ...
                ]
                "false_negatives": [
                    "groundtruth32",
                    "groundtruth24",
                    ...
                ]
            },
            "parameters": {
                "score_threshold": 0.5,
                "iou_threshold": 0.5,
            }
        }

        Parameters
        ----------
        datum_id : str
            A string identifier representing a datum.
        true_positives : list[tuple[str, str]]
            A list of string identifier pairs representing true positive ground truth and prediction combinations.
        false_positives : list[str]
            A list of string identifiers representing false positive predictions.
        false_negatives : list[str]
            A list of string identifiers representing false negative ground truths.
        score_threshold : float
            The confidence score threshold used to filter predictions.
        iou_threshold : float
            The Intersection over Union (IOU) threshold used to determine true positives.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.Examples.value,
            value={
                "datum_id": datum_id,
                "true_positives": true_positives,
                "false_positives": false_positives,
                "false_negatives": false_negatives,
            },
            parameters={
                "score_threshold": score_threshold,
                "iou_threshold": iou_threshold,
            },
        )

    @classmethod
    def confusion_matrix_with_examples(
        cls,
        confusion_matrix: dict[
            str,  # ground truth label value
            dict[
                str,  # prediction label value
                dict[
                    str,  # either `count` or `examples`
                    int
                    | list[
                        dict[
                            str,  # either `datum_id`, `ground_truth_id`, `prediction_id`
                            str,  # string identifier
                        ]
                    ],
                ],
            ],
        ],
        unmatched_predictions: dict[
            str,  # prediction label value
            dict[
                str,  # either `count` or `examples`
                int
                | list[
                    dict[
                        str,  # either `datum_id` or `prediction_id``
                        str,  # string identifier
                    ]
                ],
            ],
        ],
        unmatched_ground_truths: dict[
            str,  # ground truth label value
            dict[
                str,  # either `count` or `examples`
                int
                | list[
                    dict[
                        str,  # either `datum_id` or `ground_truth_id`
                        str,  # string identifier
                    ]
                ],
            ],
        ],
        score_threshold: float,
        iou_threshold: float,
    ):
        """
        Confusion matrix with examples for object detection tasks.

        This class encapsulates detailed information about the model's performance, including correct
        predictions, misclassifications, unmatched_predictions (subset of false positives), and unmatched ground truths
        (subset of false negatives). It provides counts and examples for each category to facilitate in-depth analysis.

        Confusion Matrix Format:
        {
            <ground truth label>: {
                <prediction label>: {
                    'count': int,
                    'examples': [
                        {
                            'datum_id': str,
                            'groundtruth_id': str,
                            'prediction_id': str
                        },
                        ...
                    ],
                },
                ...
            },
            ...
        }

        Unmatched Predictions Format:
        {
            <prediction label>: {
                'count': int,
                'examples': [
                    {
                        'datum_id': str,
                        'prediction_id': str
                    },
                    ...
                ],
            },
            ...
        }

        Unmatched Ground Truths Format:
        {
            <ground truth label>: {
                'count': int,
                'examples': [
                    {
                        'datum_id': str,
                        'groundtruth_id': str
                    },
                    ...
                ],
            },
            ...
        }

        Parameters
        ----------
        confusion_matrix : dict
            A nested dictionary where the first key is the ground truth label value, the second key
            is the prediction label value, and the innermost dictionary contains either a `count`
            or a list of `examples`. Each example includes annotation and datum identifers.
        unmatched_predictions : dict
            A dictionary where each key is a prediction label value with no corresponding ground truth
            (subset of false positives). The value is a dictionary containing either a `count` or a list of
            `examples`. Each example includes annotation and datum identifers.
        unmatched_groundtruths : dict
            A dictionary where each key is a ground truth label value for which the model failed to predict
            (subset of false negatives). The value is a dictionary containing either a `count` or a list of `examples`.
            Each example includes annotation and datum identifers.
        score_threshold : float
            The confidence score threshold used to filter predictions.
        iou_threshold : float
            The Intersection over Union (IOU) threshold used to determine true positives.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.ConfusionMatrixWithExamples.value,
            value={
                "confusion_matrix": confusion_matrix,
                "unmatched_predictions": unmatched_predictions,
                "unmatched_ground_truths": unmatched_ground_truths,
            },
            parameters={
                "score_threshold": score_threshold,
                "iou_threshold": iou_threshold,
            },
        )



from dataclasses import dataclass
from pathlib import Path

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

from valor_lite.cache import FileCacheReader, MemoryCacheReader


@dataclass
class EvaluatorInfo:
    number_of_rows: int = 0
    number_of_datums: int = 0
    number_of_labels: int = 0
    number_of_pixels: int = 0
    number_of_groundtruth_pixels: int = 0
    number_of_prediction_pixels: int = 0
    metadata_fields: list[tuple[str, str | pa.DataType]] | None = None


def generate_cache_path(path: str | Path) -> Path:
    """Generate cache path from parent directory."""
    return Path(path) / "counts"


def generate_metadata_path(path: str | Path) -> Path:
    """Generate metadata path from parent directory."""
    return Path(path) / "metadata.json"


def generate_schema(
    metadata_fields: list[tuple[str, str | pa.DataType]] | None
) -> pa.Schema:
    """Generate PyArrow schema from metadata fields."""

    metadata_fields = metadata_fields if metadata_fields else []
    reserved_fields = [
        ("datum_uid", pa.string()),
        ("datum_id", pa.int64()),
        # groundtruth
        ("gt_label", pa.string()),
        ("gt_label_id", pa.int64()),
        # prediction
        ("pd_label", pa.string()),
        ("pd_label_id", pa.int64()),
        # pair
        ("count", pa.uint64()),
    ]

    # validate
    reserved_field_names = {f[0] for f in reserved_fields}
    metadata_field_names = {f[0] for f in metadata_fields}
    if conflicting := reserved_field_names & metadata_field_names:
        raise ValueError(
            f"metadata fields {conflicting} conflict with reserved fields"
        )

    return pa.schema(
        [
            *reserved_fields,
            *metadata_fields,
        ]
    )


def encode_metadata_fields(
    metadata_fields: list[tuple[str, str | pa.DataType]] | None
) -> dict[str, str]:
    """Encode metadata fields into JSON format."""
    metadata_fields = metadata_fields if metadata_fields else []
    return {k: str(v) for k, v in metadata_fields}


def decode_metadata_fields(
    encoded_metadata_fields: dict[str, str]
) -> list[tuple[str, str | pa.DataType]]:
    """Decode metadata fields from JSON format."""
    return [(k, v) for k, v in encoded_metadata_fields.items()]


def extract_labels(
    reader: MemoryCacheReader | FileCacheReader,
    index_to_label_override: dict[int, str] | None = None,
) -> dict[int, str]:
    if index_to_label_override is not None:
        return index_to_label_override

    index_to_label = {}
    for tbl in reader.iterate_tables(
        columns=[
            "gt_label_id",
            "gt_label",
            "pd_label_id",
            "pd_label",
        ]
    ):

        # get gt labels
        gt_label_ids = tbl["gt_label_id"].to_numpy()
        gt_label_ids, gt_indices = np.unique(gt_label_ids, return_index=True)
        gt_labels = tbl["gt_label"].take(gt_indices).to_pylist()
        gt_labels = dict(zip(gt_label_ids.astype(int).tolist(), gt_labels))
        gt_labels.pop(-1, None)
        index_to_label.update(gt_labels)

        # get pd labels
        pd_label_ids = tbl["pd_label_id"].to_numpy()
        pd_label_ids, pd_indices = np.unique(pd_label_ids, return_index=True)
        pd_labels = tbl["pd_label"].take(pd_indices).to_pylist()
        pd_labels = dict(zip(pd_label_ids.astype(int).tolist(), pd_labels))
        pd_labels.pop(-1, None)
        index_to_label.update(pd_labels)

    return index_to_label


def extract_counts(
    reader: MemoryCacheReader | FileCacheReader,
    datums: pc.Expression | None = None,
    groundtruths: pc.Expression | None = None,
    predictions: pc.Expression | None = None,
):
    n_dts, n_total, n_gts, n_pds = 0, 0, 0, 0
    for tbl in reader.iterate_tables(filter=datums):

        # count datums
        n_dts += int(np.unique(tbl["datum_id"].to_numpy()).shape[0])

        # count pixels
        n_total += int(tbl["count"].to_numpy().sum())

        # count groundtruth pixels
        gt_tbl = tbl
        gt_expr = pc.field("gt_label_id") >= 0
        if groundtruths is not None:
            gt_expr &= groundtruths
        gt_tbl = tbl.filter(gt_expr)
        n_gts += int(gt_tbl["count"].to_numpy().sum())

        # count prediction pixels
        pd_tbl = tbl
        pd_expr = pc.field("pd_label_id") >= 0
        if predictions is not None:
            pd_expr &= predictions
        pd_tbl = tbl.filter(pd_expr)
        n_pds += int(pd_tbl["count"].to_numpy().sum())

    return n_dts, n_total, n_gts, n_pds



class InvalidLLMResponseError(Exception):
    """
    Raised when the response from the LLM is invalid for a given metric computation.
    """

    pass


class MismatchingTextGenerationDatumError(Exception):
    """
    Raised when datums with the same uid but different text are added to the ValorTextGenerationStreamingManager.
    """

    pass



from collections.abc import Iterator
from typing import Any

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc


class MemoryCache:
    def __init__(
        self,
        table: pa.Table,
        batch_size: int,
    ):
        self._table = table
        self._batch_size = batch_size

    @property
    def schema(self) -> pa.Schema:
        return self._table.schema

    @property
    def batch_size(self) -> int:
        return self._batch_size

    def count_tables(self) -> int:
        """Count the number of tables in the cache."""
        return 1

    def count_rows(self) -> int:
        """Count the number of rows in the cache."""
        return self._table.num_rows


class MemoryCacheReader(MemoryCache):
    def iterate_tables(
        self,
        columns: list[str] | None = None,
        filter: pc.Expression | None = None,
    ) -> Iterator[pa.Table]:
        """
        Iterate over tables within the cache.

        Parameters
        ----------
        columns : list[str], optional
            Optionally select columns to be returned.
        filter : pyarrow.compute.Expression, optional
            Optionally filter table before returning.

        Returns
        -------
        Iterator[pa.Table]
        """
        table = self._table
        if filter is not None:
            table = table.filter(filter)
        if columns is not None:
            table = table.select(columns)
        yield table

    def iterate_arrays(
        self,
        numeric_columns: list[str] | None = None,
        filter: pc.Expression | None = None,
    ) -> Iterator[np.ndarray]:
        """
        Iterate over chunks within the cache returning arrays.

        Parameters
        ----------
        numeric_columns : list[str], optional
            Optionally select numeric columns to be returned within an array.
        filter : pyarrow.compute.Expression, optional
            Optionally filter table before returning.

        Returns
        -------
        Iterator[np.ndarray]
        """
        for tbl in self.iterate_tables(columns=numeric_columns, filter=filter):
            yield np.column_stack(
                [tbl.column(i).to_numpy() for i in range(tbl.num_columns)]
            )

    def iterate_tables_with_arrays(
        self,
        columns: list[str] | None = None,
        filter: pc.Expression | None = None,
        numeric_columns: list[str] | None = None,
    ) -> Iterator[tuple[pa.Table, np.ndarray]]:
        """
        Iterate over chunks within the cache returning both tables and arrays.

        Parameters
        ----------
        columns : list[str], optional
            Optionally select columns to be returned.
        filter : pyarrow.compute.Expression, optional
            Optionally filter table before returning.
        numeric_columns : list[str], optional
            Optionally select numeric columns to be returned within an array.

        Returns
        -------
        Iterator[tuple[pa.Table, np.ndarray]]

        """
        _columns = set(columns) if columns else set()
        _numeric_columns = set(numeric_columns) if numeric_columns else set()
        columns = list(_columns.union(_numeric_columns))
        for tbl in self.iterate_tables(
            columns=columns,
            filter=filter,
        ):
            table_columns = numeric_columns if numeric_columns else tbl.columns
            yield tbl, np.column_stack(
                [tbl[col].to_numpy() for col in table_columns]
            )

    def iterate_fragments(
        self, batch_size: int
    ) -> Iterator[Iterator[pa.RecordBatch]]:
        """
        Yield a table batch iterator.

        This is intended to emulate file-based access patterns.

        Parameters
        ----------
        batch_size : int
            Maximum number of rows allowed to be read per batch.

        Yields
        ------
        Iterator[Iterator[pa.RecordBatch]]
        """
        yield iter(self._table.to_batches(max_chunksize=batch_size))


class MemoryCacheWriter(MemoryCache):
    def __init__(
        self,
        table: pa.Table,
        batch_size: int,
    ):
        super().__init__(
            table=table,
            batch_size=batch_size,
        )

        # internal state
        self._buffer = []

    @classmethod
    def create(
        cls,
        schema: pa.Schema,
        batch_size: int,
    ):
        """
        Create an in-memory cache.

        Parameters
        ----------
        schema : pa.Schema
            Cache schema.
        batch_size : int
            Target batch size when writing chunks.
        """
        return cls(
            table=schema.empty_table(),
            batch_size=batch_size,
        )

    def write_rows(
        self,
        rows: list[dict[str, Any]],
    ):
        """
        Write rows to cache.

        Parameters
        ----------
        rows : list[dict[str, Any]]
            A list of rows represented by dictionaries mapping fields to values.
        """
        if not rows:
            return
        batch = pa.RecordBatch.from_pylist(rows, schema=self.schema)
        self.write_batch(batch)

    def write_columns(
        self,
        columns: dict[str, list | np.ndarray | pa.Array],
    ):
        """
        Write columnar data to cache.

        Parameters
        ----------
        columns : dict[str, list | np.ndarray | pa.Array]
            A mapping of columnar field names to list of values.
        """
        if not columns:
            return
        batch = pa.RecordBatch.from_pydict(columns)
        self.write_batch(batch)

    def write_batch(
        self,
        batch: pa.RecordBatch,
    ):
        """
        Write a batch to cache.

        Parameters
        ----------
        batch : pa.RecordBatch
            A batch of columnar data.
        """
        size = batch.num_rows
        if self._buffer:
            size += sum([b.num_rows for b in self._buffer])

        # check size
        if size < self._batch_size:
            self._buffer.append(batch)
            return

        if self._buffer:
            self._buffer.append(batch)
            combined_arrays = [
                pa.concat_arrays([b.column(name) for b in self._buffer])
                for name in self.schema.names
            ]
            batch = pa.RecordBatch.from_arrays(
                combined_arrays, schema=self.schema
            )
            self._buffer = []

        # write batch
        self.write_table(pa.Table.from_batches([batch]))

    def write_table(
        self,
        table: pa.Table,
    ):
        """
        Write a table directly to cache.

        Parameters
        ----------
        table : pa.Table
            A populated table.
        """
        self._table = pa.concat_tables([self._table, table])

    def flush(self):
        """Flush the cache buffer."""
        if self._buffer:
            combined_arrays = [
                pa.concat_arrays([b.column(name) for b in self._buffer])
                for name in self.schema.names
            ]
            batch = pa.RecordBatch.from_arrays(
                combined_arrays, schema=self.schema
            )
            self._table = pa.concat_tables(
                [self._table, pa.Table.from_batches([batch])]
            )
            self._buffer = []

    def sort_by(
        self,
        sorting: list[tuple[str, str]],
    ):
        """
        Sort cache in-place.

        Parameters
        ----------
        sorting : list[tuple[str, str]]
            Sorting arguments in PyArrow format (e.g. [('a', 'ascending'), ('b', 'descending')]).
        """
        self.flush()
        self._table = self._table.sort_by(sorting)

    def __enter__(self):
        """Context manager entry."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit - ensures data is flushed."""
        self.flush()

    def to_reader(self) -> MemoryCacheReader:
        """Get cache reader."""
        self.flush()
        return MemoryCacheReader(
            table=self._table, batch_size=self._batch_size
        )



from .compute import sort
from .ephemeral import MemoryCacheReader, MemoryCacheWriter
from .persistent import FileCacheReader, FileCacheWriter

__all__ = [
    "FileCacheReader",
    "FileCacheWriter",
    "MemoryCacheReader",
    "MemoryCacheWriter",
    "sort",
]
